"""Firefox common functions."""

from __future__ import annotations

import itertools
import sqlite3
import time
import typing as ty
from configparser import RawConfigParser
from contextlib import closing
from pathlib import Path

from kupfer.obj import UrlLeaf
from kupfer.support import pretty

MAX_ITEMS = 10000
_BOOKMARKS_SQL = """
SELECT moz_places.url, moz_bookmarks.title
FROM moz_places, moz_bookmarks
WHERE moz_places.id = moz_bookmarks.fk
    AND moz_bookmarks.keyword_id IS NULL
ORDER BY visit_count DESC
LIMIT ?"""


def make_absolute_and_check(firefox_dir: Path, path: str) -> Path | None:
    """Helper, make path absolute and check is exist."""
    dpath = firefox_dir.joinpath(path)

    if dpath.is_dir():
        return dpath

    return None


def _find_default_profile(firefox_dir: Path) -> Path | None:
    """Try to find default/useful profile in firefox located in `firefox_dir`"""
    config = RawConfigParser({"Default": "0"})
    config.read(firefox_dir.joinpath("profiles.ini"))
    path = None

    # find Instal.* section and default profile
    for section in config.sections():
        if section.startswith("Install"):
            if not config.has_option(section, "Default"):
                continue

            # found default profile
            if path := make_absolute_and_check(
                firefox_dir, config.get(section, "Default")
            ):
                pretty.print_debug(
                    __name__, "Found install default profile", path
                )
                return path

            break

    pretty.print_debug(__name__, "Install* default profile not found")

    # not found default profile, iterate profiles, try to find default
    for section in config.sections():
        if not section.startswith("Profile"):
            continue

        if (
            config.has_option(section, "Default")
            and config.get(section, "Default") == "1"
            and (
                path := make_absolute_and_check(
                    firefox_dir, config.get(section, "Path")
                )
            )
        ):
            pretty.print_debug(
                __name__, "Found profile with default=1", section, path
            )
            return path

        # if section has path - remember it and use if default is not found
        if not path and config.has_option(section, "Path"):
            path = make_absolute_and_check(
                firefox_dir, config.get(section, "Path")
            )

    # not found default profile, return any found path (if any)
    return path


def _get_home_file(
    needed_file: str,
    profile_dir: str | Path | None,
    firefox_dir: Path,
) -> Path | None:
    """Get path to `needed_file` in `profile_dir`.

    When no `profile_dir` is not given try to find default profile
    in profiles.ini. `profile_dir` may be only profile name and is relative
    to ~/.mozilla/firefox or may be full path to profile dir.
    """
    if profile_dir:
        # user define profile name or dir, check it and if valid use id
        profile_dir = Path(profile_dir).expanduser()
        if not profile_dir.is_absolute():
            profile_dir = firefox_dir.joinpath(profile_dir)

        if not profile_dir.is_dir():
            # fail; given profile not exists
            pretty.print_debug(
                __name__, "Firefox custom profile_dir not exists", profile_dir
            )
            return None

        return profile_dir.joinpath(needed_file)

    if not firefox_dir.exists():
        pretty.print_debug(__name__, "Firefox dir not exists", firefox_dir)
        return None

    if not firefox_dir.joinpath("profiles.ini").is_file():
        pretty.print_debug(
            __name__, "Firefox profiles.ini not exists", firefox_dir
        )
        return None

    pretty.print_debug(__name__, "Firefox dir", firefox_dir)

    path = _find_default_profile(firefox_dir)
    pretty.print_debug(__name__, "Profile path", path)

    return path.joinpath(needed_file) if path else None


def get_firefox_home_file(
    needed_file: str, profile_dir: str | Path | None = None
) -> Path | None:
    """Get path to `needed_file` in `profile_dir`.

    When no `profile_dir` is not given try to find default profile
    in profiles.ini. `profile_dir` may be only profile name and is relative
    to ~/.mozilla/firefox or may be full path to profile dir.
    """
    firefox_dir = Path("~/.mozilla/firefox").expanduser()
    return _get_home_file(needed_file, profile_dir, firefox_dir)


def get_librewolf_home_file(
    needed_file: str, profile_dir: str | Path | None = None
) -> Path | None:
    """Get path to `needed_file` in `profile_dir`.

    When no `profile_dir` is not given try to find default profile
    in profiles.ini. `profile_dir` may be only profile name and is relative
    to ~/.librewolf or may be full path to profile dir.
    """
    firefox_dir = Path("~/.librewolf").expanduser()
    return _get_home_file(needed_file, profile_dir, firefox_dir)


def query_database(
    db_file_path: Path | str, sql: str, args: tuple[ty.Any, ...] = ()
) -> ty.Iterable[tuple[ty.Any, ...]]:
    """Query firefox database. Iterator must be exhausted to prevent hanging
    connection."""

    fpath = str(db_file_path).replace("?", "%3f").replace("#", "%23")
    fpath = "file:" + fpath + "?immutable=1&mode=ro"

    for _ in range(2):
        try:
            pretty.print_debug(__name__, "Query Firefox db", db_file_path, sql)
            with closing(sqlite3.connect(fpath, uri=True, timeout=1)) as conn:
                cur = conn.cursor()
                cur.execute(sql, args)
                yield from cur
                return

        except sqlite3.Error as err:
            # Something is wrong with the database
            # wait short time and try again
            pretty.print_error(__name__, "Query Firefox db error:", str(err))
            time.sleep(1)


def get_bookmarks(
    path: Path | None, max_items: int = MAX_ITEMS
) -> list[UrlLeaf]:
    """Load bookmarks from given `path` database."""

    if not path:
        return []

    return list(
        itertools.starmap(
            UrlLeaf,
            query_database(path, _BOOKMARKS_SQL, (max_items,)),
        )
    )



from __future__ import annotations

__kupfer_name__ = _("Aria2")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_actions__ = ("Download",)
__description__ = _("Download files using remote instance of Aria2")
__version__ = "2023-05-14"
__author__ = "KB"

import json
import time
import typing as ty
import urllib.request

from kupfer import plugin_support
from kupfer.obj import Action, OperationError, UrlLeaf
from kupfer.ui import uiutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "aria2_url",
        "label": _("Aria2 URL"),
        "type": str,
        "value": "",
        "tooltip": _(
            "Base URL for remote Aria2 instance ie. http://localhost:6800/"
        ),
    },
    {
        "key": "aria2_token",
        "label": _("Secret token"),
        "type": str,
        "value": "",
        "tooltip": _(
            "Enter the Aria2 RPC secret token (leave empty if "
            "authentication is not enabled) "
        ),
    },
)


class Download(Action):
    """Action to download file"""

    def __init__(self):
        Action.__init__(self, _("Remote Download"))

    def activate(self, leaf, iobj=None, ctx=None):
        aria2url = __kupfer_settings__["aria2_url"]
        if not aria2url:
            return

        params: list[str | list[str]] = []

        if token := __kupfer_settings__["aria2_token"]:
            params.append(f"token:{token}")

        params.append([str(leaf.object)])

        jsonreq = json.dumps(
            {
                "jsonrpc": "2.0",
                "id": f"kupfer-{time.time()}",
                "method": "aria2.addUri",
                "params": params,
            }
        )

        data = jsonreq.encode("ascii")

        with urllib.request.urlopen(aria2url, data) as response:
            if response.status == 200:  # noqa:PLR2004
                uiutils.show_notification("Kupfer", _("Download started"))
                return

            if res := response.read():
                err = res.decode()
            else:
                err = str(response.status)

            raise OperationError(f"Request failed: {err}")

    def item_types(self):
        yield UrlLeaf

    def get_description(self):
        return _("Download using remote instance of Aria2")



import unittest

from kupfer.support import itertools as kit


class TestTwoPartMapper(unittest.TestCase):
    def test1(self):
        def repfunc(string):
            print(string)
            self.assertEqual(len(string), 2)
            if string[0] == string[1]:
                return string.upper()

            return string

        # no changes
        self.assertEqual(
            kit.two_part_mapper("abbcdedeeffg123", repfunc), "abbcdedeeffg123"
        )
        self.assertEqual(
            kit.two_part_mapper("abbbcdedeeffg123", repfunc),
            "abBBcdedEEFFg123",
        )

    def test2(self):
        mapping = {
            "aa": "01",
            "bb": "23",
            "cc": "34",
        }

        def repfunc(string):
            return mapping.get(string)

        self.assertEqual(
            kit.two_part_mapper("abaabcdecccfag1aa", repfunc),
            "ab01bcde34cfag101",
        )
        self.assertEqual(
            kit.two_part_mapper("abbbcdedeeffg123", repfunc),
            "a23bcdedeeffg123",
        )
        self.assertEqual(
            kit.two_part_mapper("aaaaaaaaaaaaaaaa", repfunc),
            "0101010101010101",
        )
        self.assertEqual(
            kit.two_part_mapper("aabaabaabaab", repfunc), "01b01b01b01b"
        )
        self.assertEqual(
            kit.two_part_mapper("baabaabaabaab", repfunc), "b01b01b01b01b"
        )
        self.assertEqual(
            kit.two_part_mapper("baaabaaabaaabbaabb", repfunc),
            "b01ab01ab01a230123",
        )


class TestPeekfirst(unittest.TestCase):
    def test1(self):
        first, tail = kit.peekfirst([])
        self.assertIsNone(first)
        self.assertTrue(hasattr(tail, "__next__"))
        self.assertEqual([], list(tail))

        first, tail = kit.peekfirst([1, 2, 3, 4, 5])
        self.assertEqual(first, 1)
        self.assertTrue(hasattr(tail, "__next__"))
        self.assertEqual([1, 2, 3, 4, 5], list(tail))



from __future__ import annotations

__kupfer_name__ = "Free-text Queries"
__kupfer_sources__ = ()
__kupfer_text_sources__ = (
    "BasicTextSource",
    "PathTextSource",
    "URLTextSource",
)
__kupfer_actions__ = ("OpenTextUrl",)
__description__ = "Basic support for free-text queries"
__version__ = "2021.1"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os
import typing as ty
import urllib.error
import urllib.parse
import urllib.request
from contextlib import suppress
from pathlib import Path

from kupfer import launch
from kupfer.obj import FileLeaf, Leaf, OpenUrl, TextLeaf, TextSource, UrlLeaf
from kupfer.support import pretty, system
from kupfer.support.validators import is_url

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class BasicTextSource(TextSource):
    """The most basic TextSource yields one TextLeaf"""

    def __init__(self):
        TextSource.__init__(self, name=_("Text"))

    def get_text_items(self, text):
        if text:
            yield TextLeaf(text)

    def provides(self):
        yield TextLeaf


class PathTextSource(TextSource, pretty.OutputMixin):
    """Return existing full paths if typed"""

    def __init__(self):
        TextSource.__init__(self, name="Filesystem Text Matches")

    def get_rank(self):
        return 80

    def _is_local_file_url(self, url: str) -> bool:
        # Recognize file:/// or file://localhost/ or file://<local_hostname>/ URLs
        hostname = system.get_hostname()
        return url.startswith(
            ("file:///", "file://localhost/", f"file://{hostname}/")
        )

    def get_text_items(self, text: str) -> ty.Iterator[Leaf]:
        # Find directories or files
        if self._is_local_file_url(text):
            leaf = FileLeaf.from_uri(text)
            if leaf and leaf.is_valid():
                yield leaf

            return

        urlpath = Path(text)
        if not urlpath.is_absolute():
            urlpath = Path(system.get_homedir()).joinpath(urlpath)

        with suppress(OSError):
            filepath = urlpath.resolve(strict=True)
            if os.access(filepath, os.R_OK):
                yield FileLeaf(filepath)

    def provides(self):
        yield FileLeaf


class OpenTextUrl(OpenUrl):
    rank_adjust = 1

    def activate(self, leaf, iobj=None, ctx=None):
        if url := self._get_leaf_valid_url(leaf):
            launch.show_url(url)

    def item_types(self):
        yield Leaf

    def valid_for_item(self, leaf: Leaf) -> bool:
        return bool(self._get_leaf_valid_url(leaf))

    def _get_leaf_valid_url(self, leaf: Leaf) -> str | None:
        if isinstance(leaf, TextLeaf) and (url := is_url(leaf.object)):
            return url

        with suppress(AttributeError):
            if (trepr := leaf.get_text_representation()) and (  # type: ignore
                url := is_url(trepr)
            ):
                return url

        with suppress(AttributeError, NotImplementedError):
            for urepr in leaf.get_urilist_representation() or []:  # type: ignore
                if urepr and (url := is_url(urepr)):
                    return url

        return None


class URLTextSource(TextSource):
    """detect URLs and webpages"""

    def __init__(self):
        TextSource.__init__(self, name="URL Text Matches")

    def get_rank(self):
        return 75

    def get_text_items(self, text: str) -> ty.Iterator[Leaf]:
        # Only detect "perfect" URLs
        text = text.strip()
        if not text:
            return

        components = urllib.parse.urlparse(text)

        # scheme and netloc are required
        if not components.scheme or not components.netloc:
            return

        # check for any whitespaces; quick and dirty
        if len(text.split(None, 1)) != 1:
            return

        name = f"{components.netloc}{components.path}".strip("/")
        # Try to turn an URL-escaped string into a Unicode string
        name = urllib.parse.unquote(name) or text
        yield UrlLeaf(text, name=name)

    def provides(self):
        yield UrlLeaf



from __future__ import annotations

__kupfer_name__ = _("Documents")
__kupfer_sources__ = ("RecentsSource", "PlacesSource", "IgnoredApps")
__kupfer_actions__ = ("Toggle",)
__kupfer_contents__ = ("ApplicationRecentsSource",)
__description__ = _("Recently used documents and bookmarked folders")
__version__ = "2017.3"
__author__ = ""

import functools
import typing as ty
from os import path
from pathlib import Path

from gi.repository import Gio, Gtk
from xdg import BaseDirectory

from kupfer import icons, launch, plugin_support
from kupfer.obj import Action, AppLeaf, FileLeaf, Source, SourceLeaf, UrlLeaf
from kupfer.support import datatools, weaklib

if ty.TYPE_CHECKING:
    from gettext import gettext as _


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "max_days",
        "label": _("Max recent document days"),
        "type": int,
        "value": 28,
    },
    {
        "key": "check_doc_exist",
        "label": _("Show only existing documents"),
        "type": bool,
        "value": True,
    },
)

ALIASES = {
    "libreoffice": "soffice",
}

# Libreoffice doesn't separate them out, so we'll hack that in manually
SEPARATE_APPS = {
    "libreoffice": {
        ".doc": "libreoffice-writer",
        ".docx": "libreoffice-writer",
        ".odt": "libreoffice-writer",
        ".ods": "libreoffice-calc",
        ".xlsx": "libreoffice-calc",
        ".csv": "libreoffice-calc",
        ".odp": "libreoffice-impress",
        ".ppt": "libreoffice-impress",
        ".pptx": "libreoffice-impress",
        ".odg": "libreoffice-draw",
        ".odf": "libreoffice-math",
        ".mml": "libreoffice-math",
    }
}


def _file_path(uri: str) -> str | None:
    try:
        return ty.cast("str", Gio.File.new_for_uri(uri).get_path())
    except Exception:
        return None


@datatools.simple_cache
def _get(
    max_days: int, check_doc_exist: bool
) -> list[tuple[str, int, tuple[str, ...]]]:
    manager = Gtk.RecentManager.get_default()
    items = []
    for item in manager.get_items():
        if item.get_age() > max_days >= 0:
            continue

        if check_doc_exist and not item.exists():
            continue

        if item.get_private_hint():
            continue

        if not item.is_local():
            continue

        uri = item.get_uri()
        if file_path := _file_path(uri):
            apps_name = tuple(_get_app_id(item))
            items.append((file_path, item.get_modified(), apps_name))

    return items


def _get_app_id(item: Gtk.RecentInfo) -> ty.Iterator[str]:
    """Get applications id for given RecentInfo @item"""
    # some duplicates are expected but we can live with this
    for app in item.get_applications():
        # get app_id from application info
        instr = item.get_application_info(app)[0]
        app = app.lower()  # noqa:PLW2901
        yield app
        # get first word as app id
        if (aid := instr.split(None, 1)[0]) != app:
            yield aid


def _get_items(
    for_app_names: tuple[str, ...] | None = None,
) -> ty.Iterator[tuple[int, str]]:
    """Get recent items `for_app_names`.

    for_app_names: set of candidate app names, or None.
    Return iterable: tuple (modification time, file path)
    """
    max_days = __kupfer_settings__["max_days"]
    check_doc_exist = __kupfer_settings__["check_doc_exist"]
    for file_path, modified, apps in _get(max_days, check_doc_exist):
        if for_app_names:
            if not any(a in for_app_names for a in apps):
                continue

            ext = path.splitext(file_path)[1].lower()
            # check is any of app_id is in separate_apps dict, then check
            # extension of file - if is on list and matched application
            # is not @for_app_names list - skip file.
            # this allow to filter files for applications by file extension
            if any(
                sort_table.get(ext) not in for_app_names
                for app_id, sort_table in SEPARATE_APPS.items()
                if app_id in for_app_names
            ):
                continue

        yield (modified, file_path)


def _get_items_sorted(
    for_app_names: tuple[str, ...] | None = None,
) -> ty.Iterator[FileLeaf]:
    """Get recent documents as iterable FileLeaf for `for_app_names` sorted
    sorted by modified date desc."""
    items = sorted(_get_items(for_app_names), reverse=True)
    return (FileLeaf(item[1]) for item in items)


@functools.lru_cache(maxsize=10)
def _has_items_for_application(app_names: tuple[str, ...]) -> bool:
    """Check is there any recent documents for `app_names`."""
    # Cache doesn't need to be large to serve main purpose:
    # there will be many identical queries in a row
    for _item in _get_items(app_names):
        return True

    return False


def _app_names(leaf: AppLeaf) -> tuple[str, ...]:
    """Return a tuple of names for appleaf"""
    # in most cases, there are only 2-3 items, so there is not need to
    # built set
    svc = launch.get_applications_matcher_service()

    leaf_id = leaf.get_id()
    ids = [leaf_id]

    if (exe := leaf.object.get_executable()) != leaf_id:
        ids.append(exe)

    if (app_name := svc.application_name(leaf_id)) and (
        app_name := app_name.lower()
    ) != leaf_id:
        ids.append(app_name)

    ids.extend(v for k, v in ALIASES.items() if k in ids)
    # return tuple as wee need hashable object for caching
    return tuple(ids)


class RecentsSource(Source):
    def __init__(self, name=None):
        super().__init__(name or _("Recent Items"))

    def initialize(self):
        """Set up change callback"""
        manager = Gtk.RecentManager.get_default()
        weaklib.gobject_connect_weakly(
            manager, "changed", self._recent_changed
        )

    def _recent_changed(self, _rmgr: Gtk.RecentManager) -> None:
        # when using typeguard lru_cache is wrapped
        try:
            _get.cache_clear()
        except AttributeError:
            _get.__wrapped__.cache_clear()  # type:ignore

        self.mark_for_update()

    def get_items(self):
        return _get_items_sorted()

    def get_description(self):
        return _("Recently used documents")

    def get_icon_name(self):
        return "document-open-recent"

    def provides(self):
        yield FileLeaf
        yield UrlLeaf


class ApplicationRecentsSource(RecentsSource):
    def __init__(self, application):
        # TRANS: Recent Documents for application %s
        name = _("%s Documents") % str(application)
        super().__init__(name)
        self.application = application

    def repr_key(self):
        return self.application.repr_key()

    def get_items(self):
        app_names = _app_names(self.application)
        self.output_debug("Items for", app_names)
        return _get_items_sorted(app_names)

    def get_gicon(self):
        return icons.ComposedIcon(
            self.get_icon_name(), self.application.get_icon()
        )

    def get_description(self):
        return _("Recently used documents for %s") % str(self.application)

    @classmethod
    def decorates_type(cls):
        return AppLeaf

    @classmethod
    def decorate_item(cls, leaf: AppLeaf) -> ApplicationRecentsSource | None:
        if IgnoredApps.contains(leaf):
            return None

        app_names = _app_names(leaf)
        if _has_items_for_application(app_names):
            return cls(leaf)

        return None


class PlacesSource(Source):
    """Source for items from gtk bookmarks/"""

    source_scan_interval: int = 3600

    def __init__(self):
        super().__init__(_("Places"))
        self.places_file = None
        self._version = 2

    def initialize(self):
        self.places_file = path.join(
            BaseDirectory.xdg_config_home, "gtk-3.0", "bookmarks"
        )

    def get_items(self):
        """
        gtk-bookmarks: each line has url and optional title
        file:///path/to/that.end [title]
        """
        assert self.places_file
        if Path(self.places_file).exists():
            return self._get_places(self.places_file)

        return ()

    def _get_places(self, fileloc):
        with open(fileloc, encoding="UTF-8") as fin:
            for line in fin:
                line = line.strip()  # noqa:PLW2901
                if not line:
                    continue

                uri, *rest = line.split(None, 1)
                gfile = Gio.File.new_for_uri(uri)
                if rest:
                    title = rest[0]
                else:
                    disp = gfile.get_parse_name()
                    title = path.basename(disp)

                if locpath := gfile.get_path():
                    yield FileLeaf(locpath, title)
                else:
                    yield UrlLeaf(gfile.get_uri(), title)

    def get_description(self):
        return _("Bookmarked folders")

    def get_icon_name(self):
        return "system-file-manager"

    def provides(self):
        yield FileLeaf
        yield UrlLeaf


class IgnoredApps(Source):
    # This Source is invisible and has no content
    # It exists just to store (through the config mechanism) the list of apps
    # we ignore for recent documents content decoration
    instance: IgnoredApps = None  # type:ignore
    source_scan_interval: int = 3600

    def __init__(self):
        super().__init__(_("Toggle Recent Documents"))
        self._version = 2
        self.apps: set[str] = set()

    def config_save(self):
        return self.apps

    def config_save_name(self):
        return __name__

    def config_restore(self, state):
        self.apps = state

    def initialize(self):
        IgnoredApps.instance = self

    def finalize(self):
        del IgnoredApps.instance

    def get_items(self):
        return []

    def provides(self):
        return ()

    @classmethod
    def add(cls, app_leaf):
        assert cls.instance
        cls.instance.apps.add(app_leaf.get_id())
        # FIXME: Semi-hack to refresh the content
        cls.instance.mark_for_update()

    @classmethod
    def remove(cls, app_leaf):
        assert cls.instance
        cls.instance.apps.discard(app_leaf.get_id())
        cls.instance.mark_for_update()

    @classmethod
    def contains(cls, app_leaf):
        assert cls.instance
        return app_leaf.get_id() in cls.instance.apps

    def get_leaf_repr(self):
        return InvisibleSourceLeaf(self)


class Toggle(Action):
    rank_adjust = -5

    def __init__(self):
        super().__init__(_("Toggle Recent Documents"))

    def item_types(self):
        yield AppLeaf

    def valid_for_item(self, leaf):
        if IgnoredApps.contains(leaf):
            return True

        app_names = _app_names(leaf)
        return _has_items_for_application(app_names)

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        if IgnoredApps.contains(leaf):
            IgnoredApps.remove(leaf)
        else:
            IgnoredApps.add(leaf)
        # Neat trick: We return the leaf, and that updates the decoration
        # pylint: disable=protected-access
        leaf._content_source = None  # noqa: SLF001
        return leaf

    def get_description(self):
        return _(
            "Enable/disable listing recent documents in "
            "content for this application"
        )


class InvisibleSourceLeaf(SourceLeaf):
    """Hack to hide this source"""

    def is_valid(self):
        return False



#! /usr/bin/env python3
from __future__ import annotations

import enum
import itertools
import typing as ty

from gi.repository import Gdk, GdkPixbuf, GLib, GObject, Gtk

import kupfer.config
import kupfer.environment
from kupfer import icons
from kupfer.core import actionaccel, learn, relevance, search, settings
from kupfer.obj import Action, AnySource, KupferObject, Leaf
from kupfer.support import pretty
from kupfer.ui._support import escape_markup_str, text_direction_is_ltr

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.core.search import Rankable


def _format_match(match: str) -> str:
    """Function used to format matched text in label."""
    return f"<u><b>{escape_markup_str(match)}</b></u>"


# State Constants
class State(enum.IntEnum):
    WAIT = 1
    MATCH = 2
    NO_MATCH = 3


_PREEDIT_HIDDEN_CLASS: ty.Final = "hidden"
_WINDOW_BORDER_WIDTH: ty.Final = 8

# columns position
_ICON_COL: ty.Final = 1
_NAME_COL: ty.Final = 2
_FAV_COL: ty.Final = 3
_INFO_COL: ty.Final = 4
_RANK_COL: ty.Final = 5

_MIN_ICON_SIZE_TO_SHOW: ty.Final[int] = 8


class _LeafModel:
    """A base for a tree view with a magic load-on-demand feature.

    self.set_base will set its base iterator and self.populate(num) will load
    @num items into the model

    Attributes:
    icon_size
    """

    def __init__(
        self, aux_info_callback: ty.Callable[[search.RankableObject], str]
    ) -> None:
        """First column is always the object -- returned by get_object
        it needs not be specified in columns."""
        self.icon_size = 32
        columns = (GObject.TYPE_OBJECT, str, str, str, str)
        self._store = Gtk.ListStore(GObject.TYPE_PYOBJECT, *columns)
        self._base: ty.Iterator[Rankable] | None = None
        self._setup_columns()
        self._aux_info_callback = aux_info_callback

    def __len__(self) -> int:
        return len(self._store)

    def _setup_columns(self):
        # Name and description column
        # Expands to the rest of the space
        name_cell = Gtk.CellRendererText()
        setctl = settings.get_settings_controller()
        name_cell.set_property("ellipsize", setctl.get_ellipsize_mode())
        name_col = Gtk.TreeViewColumn("item", name_cell)
        name_col.set_expand(True)
        name_col.add_attribute(name_cell, "markup", _NAME_COL)

        fav_cell = Gtk.CellRendererText()
        fav_col = Gtk.TreeViewColumn("fav", fav_cell)
        fav_col.add_attribute(fav_cell, "text", _FAV_COL)

        info_cell = Gtk.CellRendererText()
        info_col = Gtk.TreeViewColumn("info", info_cell)
        info_col.add_attribute(info_cell, "text", _INFO_COL)

        nbr_cell = Gtk.CellRendererText()
        nbr_col = Gtk.TreeViewColumn("rank", nbr_cell)
        nbr_cell.set_property("width-chars", 3)
        nbr_col.add_attribute(nbr_cell, "text", _RANK_COL)

        icon_cell = Gtk.CellRendererPixbuf()
        icon_col = Gtk.TreeViewColumn("icon", icon_cell)
        icon_col.add_attribute(icon_cell, "pixbuf", _ICON_COL)

        self.columns = [icon_col, name_col, fav_col, info_col]

        # show rank only in debug mode
        if pretty.DEBUG:
            self.columns.append(nbr_col)

    def get_object(self, path: ty.Iterable[int] | None) -> Rankable | None:
        """Get object for given `path`."""
        if path is None:
            return None

        store_iter = self._store.get_iter(path)
        return self._store.get_value(store_iter, 0)  # type: ignore

    def get_store(self) -> Gtk.ListStore:
        """Get list store."""
        return self._store

    def clear(self) -> None:
        """Clear the model and reset its base"""
        self._store.clear()
        self._base = None

    def set_base(self, baseiter: ty.Iterable[Rankable]) -> None:
        """Set base iterator that provide items for model."""
        self._base = iter(baseiter)

    def populate(self, num: int | None = None) -> KupferObject | None:
        """Populate model with num items from its base and return first item
        inserted. If num is none, insert everything.
        """
        if not self._base:
            return None

        iterator: ty.Iterator[Rankable] = self._base
        if num:
            iterator = itertools.islice(self._base, num)

        append = self._store.append
        build_row = self._build_row

        try:
            first_rank = next(iterator)
            append(build_row(first_rank))
            first = first_rank.object
        except StopIteration:
            return None

        for item in iterator:
            append(build_row(item))

        # first.object is a leaf
        return first

    def _build_row(
        self, rankable: Rankable
    ) -> tuple[Rankable, GdkPixbuf.Pixbuf | None, str, str, str, str]:
        """Use the UI description functions get_* to initialize `rankable` into
        the model. Return (rankable, icon, markup, fav, info, rank_str).
        """
        leaf, rank = rankable.object, rankable.rank
        assert isinstance(leaf, (Leaf, Action))
        return (
            rankable,
            self._get_icon(leaf),
            self._get_label_markup(leaf),
            self._get_fav(leaf),
            self._get_aux_info(leaf),
            self._get_rank_str(rank),
        )

    def _iter_store(self) -> ty.Iterator[tuple[Rankable, Gtk.TreeIter]]:
        """Iterate over item in tree store and yield rankable object and iterator"""
        siter = self._store.get_iter_first()
        while siter:
            yield self._store.get(siter, 0)[0], siter
            siter = self._store.iter_next(siter)

    def add_first(self, rankable: Rankable) -> None:
        """Add rankable on top the list. Remove previous object when already
        exists.
        """
        # first check is object already exists
        for row_rankable, siter in self._iter_store():
            if row_rankable.object == rankable.object:
                # object already on list; remove it
                self._store.remove(siter)
                break

        self._store.prepend(self._build_row(rankable))

    def find(self, obj: search.RankableObject) -> int:
        """Find @obj in store and return it row number"""
        for row, (row_rankable, _siter) in enumerate(self._iter_store()):
            if row_rankable.object == obj:
                return row

        return -1

    def _get_icon(
        self, leaf: search.RankableObject
    ) -> GdkPixbuf.Pixbuf | None:
        """Get icon from `leaf` to show in row."""
        if (size := self.icon_size) > _MIN_ICON_SIZE_TO_SHOW:
            return leaf.get_thumbnail(size, size) or leaf.get_pixbuf(size)

        return None

    def _get_label_markup(self, leaf: search.RankableObject) -> str:
        """Get `rankable` description to show in row."""
        # Here we use the items real name.
        # Previously we used the alias that was matched, but it can be too
        # confusing or ugly
        name = escape_markup_str(str(leaf))
        if desc := escape_markup_str(leaf.get_description() or ""):
            return f"{name}\n<small>{desc}</small>"

        return name

    def _get_fav(self, leaf: search.RankableObject) -> str:
        """Get star if `leaf` is favourite."""
        if learn.is_favorite(leaf):
            return "\N{BLACK STAR}"

        return ""

    def _get_aux_info(self, leaf: search.RankableObject) -> str:
        """Show additional information about leaves using aux_info_callback.
        For objects: Show arrow if it has content.
        For actions: Show accelerator.
        """
        if self._aux_info_callback is not None:
            return self._aux_info_callback(leaf)

        return ""

    def _get_rank_str(self, rank: float | None) -> str:
        # Display rank empty instead of 0 since it looks better
        return str(int(rank)) if rank else ""


def _dim_icon(icon: GdkPixbuf.Pixbuf | None) -> GdkPixbuf.Pixbuf | None:
    if not icon:
        return None

    dim_icon = icon.copy()
    dim_icon.fill(0)
    icon.composite(
        dim_icon,
        0,
        0,
        icon.get_width(),
        icon.get_height(),
        0,
        0,
        1.0,
        1.0,
        GdkPixbuf.InterpType.NEAREST,
        127,
    )
    return dim_icon


_LABEL_CHAR_WIDTH: ty.Final = 25
_PREEDIT_CHAR_WIDTH: ty.Final = 5


# pylint: disable=too-many-instance-attributes
class MatchViewOwner(pretty.OutputMixin):
    """Owner of the widget for displaying name, icon and name underlining (if
    applicable) of the current match."""

    def __init__(self) -> None:
        # object attributes
        self._match_state: State = State.WAIT

        self.object_stack: list[Leaf] = []

        # finally build widget
        self._build_widget()

        self._cur_icon: GdkPixbuf.Pixbuf | None = None
        self._cur_text: str | None = None
        self._cur_match: str | None = None
        self._icon_size: int = 0

        self._read_icon_size()

    def _on_icon_size_changed(
        self,
        setctl: settings.SettingsController,
        section: str | None,
        key: str | None,
        value: ty.Any,
    ) -> None:
        self._icon_size = setctl.get_config_int(
            "Appearance", "icon_large_size"
        )

    def _read_icon_size(self, *_args: ty.Any) -> None:
        setctl = settings.get_settings_controller()
        setctl.connect(
            "value-changed::appearance.icon_large_size",
            self._on_icon_size_changed,
        )
        self._on_icon_size_changed(setctl, None, None, None)

    def _build_widget(self) -> None:
        """Core initialization method that builds the widget."""
        self._label = label = Gtk.Label.new("<match>")
        label.set_single_line_mode(True)
        label.set_width_chars(_LABEL_CHAR_WIDTH)
        label.set_max_width_chars(_LABEL_CHAR_WIDTH)

        setctl = settings.get_settings_controller()
        label.set_ellipsize(setctl.get_ellipsize_mode())

        self._icon_view = Gtk.Image()

        # infobox: icon and match name
        icon_align = Gtk.Alignment.new(0.5, 0.5, 0, 0)
        icon_align.set_property("top-padding", 5)
        icon_align.add(self._icon_view)

        infobox = Gtk.HBox()
        infobox.pack_start(icon_align, True, True, 0)

        box = Gtk.VBox()
        box.pack_start(infobox, True, False, 0)

        self._editbox = Gtk.HBox()
        self._editbox.pack_start(self._label, True, True, 0)
        box.pack_start(self._editbox, False, False, 3)

        event_box = Gtk.EventBox()
        event_box.add(box)
        event_box.get_style_context().add_class("matchview")
        event_box.show_all()
        self._child = event_box

    def widget(self) -> Gtk.Widget:
        """Return the corresponding Widget"""
        return self._child

    # pylint: disable=too-many-locals
    def _render_composed_icon(
        self,
        base: GdkPixbuf.Pixbuf,
        pixbufs: list[GdkPixbuf.Pixbuf],
        small_size: int,
    ) -> GdkPixbuf.Pixbuf:
        """Render the main selection + a string of objects on the stack.

        Scale the main image into the upper portion, leaving a clear
        strip at the bottom where we line up the small icons.

        @base: main selection pixbuf
        @pixbufs: icons of the object stack, in final (small) size
        @small_size: the size of the small icons
        """
        size = self._icon_size
        assert size
        base_scale = min(
            (size - small_size) * 1.0 / base.get_height(),
            size * 1.0 / base.get_width(),
        )
        new_sz_x = int(base_scale * base.get_width())
        new_sz_y = int(base_scale * base.get_height())
        if not base.get_has_alpha():
            base = base.add_alpha(False, 0, 0, 0)

        destbuf = base.scale_simple(size, size, GdkPixbuf.InterpType.NEAREST)
        destbuf.fill(0x00000000)
        # Align in the middle of the area
        offset_x = (size - new_sz_x) / 2
        offset_y = ((size - small_size) - new_sz_y) / 2
        base.composite(
            destbuf,
            offset_x,
            offset_y,
            new_sz_x,
            new_sz_y,
            offset_x,
            offset_y,
            base_scale,
            base_scale,
            GdkPixbuf.InterpType.BILINEAR,
            255,
        )

        # @fr is the scale compared to the destination pixbuf
        frac = small_size * 1.0 / size
        dest_y = offset_y = int((1 - frac) * size)
        n_small = size // small_size
        for idx, pbuf in enumerate(pixbufs[-n_small:]):
            dest_x = offset_x = int(frac * size) * idx
            pbuf.copy_area(
                0, 0, small_size, small_size, destbuf, dest_x, dest_y
            )

        return destbuf

    def update_match(self) -> None:
        """Update interface to display the currently selected match."""
        # update icon
        if icon := self._cur_icon:
            if self._match_state is State.NO_MATCH:
                icon = _dim_icon(icon)

            if icon and self.object_stack:
                small_size = small_max = 16
                pixbufs = [
                    o.get_pixbuf(small_size)
                    for o in self.object_stack[-small_max:]
                ]
                icon = self._render_composed_icon(icon, pixbufs, small_size)

            self._icon_view.set_from_pixbuf(icon)

        else:
            self._icon_view.clear()
            self._icon_view.set_pixel_size(self._icon_size)

        if not self._cur_text:
            self._label.set_text("")
            return

        if not self._cur_match:
            if self._match_state is not State.MATCH:
                # Allow markup in the text string if we have no match
                self._label.set_markup(self._cur_text)
            else:
                self._label.set_text(self._cur_text)

            return

        # update the text label
        markup = relevance.format_common_substrings(
            str(self._cur_text),  # text
            str(self._cur_match).lower(),  # key,
            format_clean=escape_markup_str,
            format_match=_format_match,
        )
        self._label.set_markup(markup)

    def set_object(
        self,
        text: str | None,
        icon: GdkPixbuf.Pixbuf | None,
        update: bool = True,
    ) -> None:
        self._cur_text = text
        self._cur_icon = icon
        if update:
            self.update_match()

    def set_match(
        self,
        match: str | None = None,
        state: State | None = None,
        update: bool = True,
    ) -> None:
        self._cur_match = match
        if state:
            self._match_state = state
        elif match is not None:
            self._match_state = State.MATCH
        else:
            self._match_state = State.NO_MATCH

        if update:
            self.update_match()

    def set_match_state(
        self,
        text: str | None,
        icon: GdkPixbuf.Pixbuf | None,
        match: str | None = None,
        state: State | None = None,
        update: bool = True,
    ) -> None:
        self.set_object(text, icon, update=False)
        self.set_match(match, state, update=False)
        if update:
            self.update_match()

    def set_match_text(self, text: str | None, update: bool = True) -> None:
        self._cur_match = text
        if update:
            self.update_match()

    def expand_preedit(self, preedit: Gtk.Entry) -> None:
        new_label_width = _LABEL_CHAR_WIDTH - _PREEDIT_CHAR_WIDTH
        self._label.set_width_chars(new_label_width)
        preedit.set_width_chars(_PREEDIT_CHAR_WIDTH)
        preedit.get_style_context().remove_class(_PREEDIT_HIDDEN_CLASS)

    def shrink_preedit(self, preedit: Gtk.Entry) -> None:
        self._label.set_width_chars(_LABEL_CHAR_WIDTH)
        preedit.set_width_chars(0)
        preedit.get_style_context().add_class(_PREEDIT_HIDDEN_CLASS)

    def inject_preedit(self, preedit: Gtk.Entry | None) -> None:
        """@preedit: Widget to be injected or None"""
        if not preedit:
            self._label.set_width_chars(_LABEL_CHAR_WIDTH)
            self._label.set_alignment(0.5, 0.5)
            return

        if old_parent := preedit.get_parent():
            old_parent.remove(preedit)

        self.shrink_preedit(preedit)
        self._editbox.pack_start(preedit, False, True, 0)
        # selectedc = self.style.dark[Gtk.StateType.SELECTED]
        # preedit.modify_bg(Gtk.StateType.SELECTED, selectedc)
        preedit.show()
        preedit.grab_focus()


# number rows to skip when press PgUp/PgDown
_PAGE_STEP: ty.Final = 7
_SHOW_MORE: ty.Final = 10


# pylint: disable=too-many-public-methods
class Search(GObject.GObject, pretty.OutputMixin):  # type:ignore
    """Owner of a widget for displaying search results (using match view),
    keeping current search result list and its display.

    Signals
    * cursor-changed: def callback(widget, selection)
        called with new selected (represented) object or None
    * activate: def callback(widget, selection)
        called with activated leaf, when the widget is activated
        by double-click in table
    * table-event: def callback(widget, table, event)
        called when the user types in the table
    """

    # minimal length of list is MULT * icon size small
    LIST_MIN_MULT = 8
    __gtype_name__ = "Search"

    def __init__(self):
        GObject.GObject.__init__(self)
        # object attributes
        self._model = _LeafModel(self._get_aux_info)
        self._match = None
        self._match_state = State.WAIT
        self._text: str | None = ""
        self._source: AnySource | None = None
        self._old_win_position: tuple[int, int] | None = None
        self._old_win_size: tuple[int, int] | None = None
        self._has_search_result = False
        self._initialized = False
        self._icon_size: int = 0
        self._icon_size_small: int = 0
        # finally build widget
        self._build_widget()
        self._read_icon_size()
        self._setup_empty()

    def _get_aux_info(self, leaf: KupferObject) -> str:
        # Return content for the aux info column
        return ""

    def set_name(self, name: str) -> None:
        """Set the `name` of the Search's widget."""
        self._child.set_name(name)

    def set_state(self, state: Gtk.StateType) -> None:
        self._child.set_state(state)

    def show(self) -> None:
        self._child.show()

    def hide(self) -> None:
        self._child.hide()

    def set_visible(self, flag: bool) -> None:
        if flag:
            self.show()
        else:
            self.hide()

    def _on_icon_size_changed(
        self,
        setctl: settings.SettingsController,
        section: str | None,
        key: str | None,
        value: ty.Any,
    ) -> None:
        self._icon_size = setctl.get_config_int(
            "Appearance", "icon_large_size"
        )
        self._icon_size_small = setctl.get_config_int(
            "Appearance", "icon_small_size"
        )
        self._model.icon_size = self._icon_size_small

    def _read_icon_size(self, *args: ty.Any) -> None:
        setctl = settings.get_settings_controller()
        setctl.connect(
            "value-changed::appearance.icon_large_size",
            self._on_icon_size_changed,
        )
        setctl.connect(
            "value-changed::appearance.icon_small_size",
            self._on_icon_size_changed,
        )
        self._on_icon_size_changed(setctl, None, None, None)

    def _build_widget(self) -> None:
        """
        Core initialization method that builds the widget
        """
        self.match_view = MatchViewOwner()

        self._table = table = Gtk.TreeView.new_with_model(
            self._model.get_store()
        )
        table.set_name("kupfer-list-view")
        table.set_headers_visible(False)
        table.set_property("enable-search", False)
        table.set_has_tooltip(True)
        table.set_tooltip_column(2)

        for col in self._model.columns:
            table.append_column(col)

        table.connect("row-activated", self._on_row_activated)
        table.connect("cursor-changed", self._on_cursor_changed)

        scroller = Gtk.ScrolledWindow()
        scroller.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scroller.add(table)  # pylint: disable=no-member
        vscroll = scroller.get_vscrollbar()
        vscroll.connect("change-value", self._on_table_scroll_changed)

        self._list_window = Gtk.Window.new(Gtk.WindowType.POPUP)
        self._list_window.set_name("kupfer-list")

        self._list_window.add(scroller)
        scroller.show_all()  # pylint: disable=no-member
        self._child = self.match_view.widget()

    def widget(self) -> Gtk.Widget:
        """Return the corresponding Widget."""
        return self._child

    def get_current(self) -> KupferObject | None:
        """return current selection"""
        return self._match

    def set_object_stack(self, stack: list[Leaf]) -> None:
        self.match_view.object_stack = stack
        self.match_view.update_match()

    def set_source(self, source: AnySource) -> None:
        """Set current source (to get icon, name etc)"""
        self._source = source

    def get_match_state(self) -> State:
        return self._match_state

    def get_match_text(self) -> str | None:
        return self._text

    def get_table_visible(self) -> bool:
        return self._list_window.get_property("visible")  # type: ignore

    def hide_table(self) -> None:
        if self.get_table_visible():
            self._list_window.hide()

    def _reposition_list_window(
        self,
    ) -> None:  # pylint: disable=too-many-locals
        widget = self.widget()
        window = widget.get_toplevel()
        win_size = window.get_size()
        if not win_size:
            return

        win_pos = window.get_position()

        # do nothing if position and size not changed
        if (
            self._old_win_position == win_pos
            and self._old_win_size == win_size
        ):
            return

        ssize = widget.translate_coordinates(window, 0, 0)
        if not ssize:
            return

        pos_x, pos_y = win_pos
        self_x = ssize[0]
        self_width = widget.size_request().width
        self_end = self_x + self_width

        sub_y = pos_y + win_size[1]
        self._table.size_request()

        setctl = settings.get_settings_controller()
        subwin_height = setctl.get_config_int("Appearance", "list_height")
        if subwin_height < self._icon_size_small * self.LIST_MIN_MULT:
            subwin_height = self.LIST_MIN_MULT * self._icon_size_small

        subwin_width = self_width * 2 + _WINDOW_BORDER_WIDTH

        if self_end < subwin_width:
            # Place snugly to left
            sub_x = pos_x + self_x
        else:
            # Place aligned with right side of window
            sub_x = pos_x + self_end - subwin_width

        self._list_window.move(sub_x, sub_y)
        self._list_window.resize(subwin_width, subwin_height)
        self._list_window.set_transient_for(window)
        self._list_window.set_property("focus-on-map", False)
        self._old_win_position = win_pos
        self._old_win_size = win_size

    def _show_table(self) -> None:  # pylint: disable=too-many-locals
        self._reposition_list_window()
        self._list_window.show()

    def show_table(self) -> None:
        self.go_down(True)

    def show_table_quirk(self) -> None:
        "Show table after being hidden in the same event"
        # KWin bugs out if we hide and show the table during the same gtk event
        # issue #47
        if kupfer.environment.is_kwin():
            GLib.idle_add(self.show_table)
        else:
            self.show_table()

    def _on_table_scroll_changed(
        self, scrollbar: Gtk.Scrollbar, _scroll_type: ty.Any, value: int
    ) -> None:
        """When the scrollbar changes due to user interaction"""
        # page size: size of currently visible area
        adj = scrollbar.get_adjustment()
        upper = adj.get_property("upper")
        page_size = adj.get_property("page-size")

        if value + page_size >= upper:
            self._populate(_SHOW_MORE)

    # table methods
    def _table_set_cursor_at_row(self, row: int) -> None:
        self._table.set_cursor((row,))
        # scroll with minimal delay
        GLib.timeout_add(100, lambda: self._table.scroll_to_cell((row,)))

    def _table_current_row(self) -> int | None:
        path, _col = self._table.get_cursor()
        return path[0] if path else None

    def go_up(self, rows_count: int = 1) -> None:
        """Upwards in the table"""
        # go up, simply. close table if we go up from row 0
        path, _col = self._table.get_cursor()
        if not path:
            return

        if (row := path[0]) >= 1:
            self._table_set_cursor_at_row(row - min(rows_count, row))
        else:
            self.hide_table()

    def go_down(
        self, force: bool = False, rows_count: int = 1, show_table: bool = True
    ) -> None:
        """Down in the table."""
        table_visible = self.get_table_visible()
        # if no data is loaded (frex viewing catalog), load
        # if too little data is loaded, try load more
        if len(self._model) <= 1:
            self._populate(_SHOW_MORE)

        if len(self._model) >= 1:
            path, _col = self._table.get_cursor()
            if path:
                row = path[0]
                if len(self._model) - rows_count <= row:
                    self._populate(_SHOW_MORE)
                # go down only if table is visible
                if table_visible and (
                    step := min(len(self._model) - row - 1, rows_count)
                ):
                    self._table_set_cursor_at_row(row + step)
            else:
                self._table_set_cursor_at_row(0)

            if show_table:
                self._show_table()

        if force and show_table:
            self._show_table()

    def go_page_up(self) -> None:
        """move list one page up"""
        self.go_up(_PAGE_STEP)

    def go_page_down(self) -> None:
        """move list one page down"""
        self.go_down(rows_count=_PAGE_STEP)

    def go_first(self) -> None:
        """Rewind to first item"""
        if self.get_table_visible():
            self._table_set_cursor_at_row(0)

    def window_config(
        self, widget: Gtk.Widget, event: Gdk.EventConfigure
    ) -> None:
        """When the window moves"""
        if self.get_table_visible():
            self._reposition_list_window()

    def window_hidden(self, window: Gtk.Widget) -> None:
        """Window changed hid"""
        self.hide_table()

    def _on_row_activated(
        self, treeview: Gtk.TreeView, path: ty.Any, col: ty.Any
    ) -> None:
        obj = self.get_current()
        self.emit("activate", obj)

    def _on_cursor_changed(self, treeview: Gtk.TreeView) -> None:
        path, _col = treeview.get_cursor()
        match = self._model.get_object(path)
        self._set_match(match)

    def _set_match(self, rankable: Rankable | None = None) -> None:
        """Set the currently selected (represented) object, either as
        @rankable or KupferObject @obj

        Emits cursor-changed
        """
        self._match = rankable.object if rankable else None
        self.emit("cursor-changed", self._match)
        if self._match:
            match_text = rankable.value if rankable else None
            self._match_state = State.MATCH
            icon_size = self._icon_size
            pbuf = self._match.get_thumbnail(
                icon_size * 4 // 3, icon_size
            ) or self._match.get_pixbuf(icon_size)
            self.match_view.set_match_state(
                match_text, pbuf, match=self._text, state=self._match_state
            )

    def set_match_plain(self, obj: Rankable) -> None:
        """Set match to object @obj, without search or matches"""
        self._text = None
        self._set_match(obj)
        self._model.add_first(obj)
        self._table_set_cursor_at_row(0)

    def set_match_leaf(self, leaf: Leaf) -> None:
        """Set match to @leaf if it is on list"""
        if len(self._model) == 0:
            return

        while True:
            # try find
            if (row := self._model.find(leaf)) >= 0:
                self._show_table()
                self._table_set_cursor_at_row(row)
                return

            # if not found try to load more items
            if not self._populate(_SHOW_MORE):
                # stop when no more leaves
                return

    def relax_match(self) -> None:
        """Remove match text highlight"""
        self.match_view.set_match_text(None)
        self._text = None

    def has_result(self) -> bool:
        """A search with explicit search term is active"""
        return self._has_search_result

    def is_showing_result(self) -> bool:
        """Showing search result:
        A search with explicit search term is active,
        and the result list is shown.
        """
        return self._has_search_result and self.get_table_visible()

    def update_match(
        self,
        key: str | None,
        matchrankable: Rankable | None,
        matches: ty.Iterable[Rankable],
    ) -> None:
        """
        @matchrankable: Rankable first match or None
        @matches: Iterable to rest of matches
        """
        self._has_search_result = bool(key)
        self._model.clear()
        self._text = key
        if not matchrankable:
            self._set_match(None)
            self._handle_no_matches(empty=not key)
            return

        self._set_match(matchrankable)
        self._model.set_base(iter(matches))
        if not self._model and self.get_table_visible():
            self.go_down()

    def reset(self) -> None:
        self._has_search_result = False
        self._initialized = True
        self._model.clear()
        self._setup_empty()

    def _setup_empty(self) -> None:
        self._match_state = State.NO_MATCH
        self.match_view.set_match_state("No match", None, state=State.NO_MATCH)
        self.relax_match()

    def _populate(self, num: int) -> KupferObject | None:
        """populate model with num items"""
        return self._model.populate(num)

    def _handle_no_matches(self, empty: bool = False) -> None:
        """if @empty, there were no matches to find"""
        assert hasattr(self, "_get_nomatch_name_icon")
        name, icon = self._get_nomatch_name_icon(  # pylint: disable=no-member
            empty=empty
        )
        self._match_state = State.NO_MATCH
        self.match_view.set_match_state(name, icon, state=State.NO_MATCH)


# Take care of GObject things to set up the Search class
GObject.type_register(Search)
GObject.signal_new(
    "activate",
    Search,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_PYOBJECT,),
)
GObject.signal_new(
    "cursor-changed",
    Search,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_PYOBJECT,),
)


class LeafSearch(Search):
    """Customize for leaves search"""

    def __init__(self):
        super().__init__()
        if text_direction_is_ltr():
            self._aux_info_str = "\N{BLACK RIGHT-POINTING SMALL TRIANGLE} "
        else:
            self._aux_info_str = "\N{BLACK LEFT-POINTING SMALL TRIANGLE} "

    def _get_aux_info(self, leaf: KupferObject) -> str:
        if hasattr(leaf, "has_content") and leaf.has_content():
            return self._aux_info_str

        return ""

    def _get_pbuf(self, src: AnySource) -> GdkPixbuf.Pixbuf:
        icon_size = self._icon_size
        return src.get_thumbnail(
            icon_size * 4 // 3, icon_size
        ) or src.get_pixbuf(icon_size)

    def _get_nomatch_name_icon(
        self, empty: bool
    ) -> tuple[str, GdkPixbuf.Pixbuf]:
        if empty and self._source:
            return (
                f"<i>{escape_markup_str(self._source.get_empty_text())}</i>",
                self._get_pbuf(self._source),
            )

        if self._source:
            assert self._text
            return (
                _('No matches in %(src)s for "%(query)s"')
                % {
                    "src": f"<i>{escape_markup_str(str(self._source))}</i>",
                    "query": escape_markup_str(self._text),
                },
                self._get_pbuf(self._source),
            )

        return _("No matches"), icons.get_icon_for_name(
            "kupfer-object", self._icon_size
        )

    def _setup_empty(self) -> None:
        if self._source:
            icon = self._get_pbuf(self._source)
            msg = self._source.get_search_text()
        else:
            icon = None
            msg = _("Type to search")

        title = f"<i>{msg}</i>"

        self._set_match(None)
        self._match_state = State.WAIT
        self.match_view.set_match_state(title, icon, state=State.WAIT)


def _accel_for_action(
    action: Action, action_accel_config: actionaccel.AccelConfig | None
) -> str | None:
    if action_accel_config is None:
        return None

    if (config_accel := action_accel_config.get(action)) is not None:
        return config_accel

    return action.action_accelerator


class ActionSearch(Search):
    """Customization for Actions

    Attributes: accel_modifier
    """

    def __init__(self) -> None:
        super().__init__()
        self.action_accel_config: actionaccel.AccelConfig | None = None
        # pylint: disable=no-member
        self.accel_modifier: Gdk.ModifierType = Gdk.ModifierType.MOD1_MASK

    def lazy_setup(self) -> None:
        setctl = settings.get_settings_controller()
        setctl.connect(
            "value-changed::kupfer.action_accelerator_modifer",
            self._on_modifier_changed,
        )
        self._read_accel_modifier(setctl.get_action_accelerator_modifer())

    def _on_modifier_changed(
        self,
        setctl: settings.SettingsController,
        section: ty.Any,
        key: ty.Any,
        value: str,
    ) -> None:
        self._read_accel_modifier(value)

    def _read_accel_modifier(self, value: str) -> None:
        if value == "alt":
            # pylint: disable=no-member
            self.accel_modifier = Gdk.ModifierType.MOD1_MASK
        elif value == "ctrl":
            # pylint: disable=no-member
            self.accel_modifier = Gdk.ModifierType.CONTROL_MASK
        else:
            pretty.print_error("Unknown modifier key", value)

    def _get_aux_info(self, leaf: Action) -> str:  # type:ignore
        if not self.action_accel_config:
            return ""

        if accel := _accel_for_action(leaf, self.action_accel_config):
            keyv, mods = Gtk.accelerator_parse(accel)
            if mods != 0:
                self.output_error("Ignoring action accelerator mod", mods)

            return Gtk.accelerator_get_label(  # type:ignore
                keyv, self.accel_modifier
            )

        return ""

    def _get_nomatch_name_icon(
        self, empty: bool = False
    ) -> tuple[str, GdkPixbuf.Pixbuf | None]:
        # don't look up icons too early
        if not self._initialized:
            return ("", None)

        if self._text:
            msg = _('No action matches "%s"') % escape_markup_str(self._text)
            title = f"<i>{msg}</i>"
        else:
            title = ""

        return title, icons.get_icon_for_name(
            "kupfer-execute", self._icon_size
        )

    def _setup_empty(self) -> None:
        self._handle_no_matches()
        self.hide_table()

    def select_action_by_accel(self, accel: str) -> tuple[bool, bool]:
        """Find and select the next action with accelerator key @accel.

        When exists two or more actions with the same accelerator, select next
        and not execute it. This allow to iterate between action by pressing
        accelerator key.

        Return pair of bool success, can activate
        """
        if self.get_match_state() == State.NO_MATCH:
            return False, False

        idx = self._table_current_row() or 0
        self._populate(1)
        if not self._model:
            return False, False

        start_row = idx
        # keep info about first found action with given accelerator (idx, action)
        matched_action = None

        while True:
            self._populate(1)
            idx = (idx + 1) % len(self._model)

            cur = self._model.get_object((idx,))
            assert cur and isinstance(cur.object, Action)
            action: Action = cur.object
            self.output_debug("Looking at action", action)

            if _accel_for_action(action, self.action_accel_config) == accel:
                if matched_action is None:
                    matched_action = (idx, action)
                else:
                    # found another action with the same accelerator
                    # select first action and exit (do not execute action)
                    self._table_set_cursor_at_row(matched_action[0])
                    return True, False

            if idx == start_row:
                break

        if matched_action:
            idx, action = matched_action
            self._table_set_cursor_at_row(idx)
            return True, not action.requires_object()

        return False, False



from __future__ import annotations

__kupfer_name__ = _("SSH Hosts")
__description__ = _("Adds the SSH hosts found in ~/.ssh/config.")
__version__ = "2023-11-04"
__author__ = "Fabian Carlström, Karol Będkowski"

__kupfer_sources__ = ("SSHSource",)
__kupfer_actions__ = ("SSHConnect", "ScpFile")

import itertools
import os
import typing as ty
from contextlib import suppress

from kupfer import icons, launch
from kupfer.obj import Action, FileLeaf
from kupfer.obj.grouping import ToplevelGroupingSource
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.obj.hosts import (
    HOST_ADDRESS_KEY,
    HOST_NAME_KEY,
    HOST_SERVICE_NAME_KEY,
    HOST_SERVICE_USER_KEY,
    HostLeaf,
)

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class SSHLeaf(HostLeaf):
    """The SSH host. Using name defined in `Host` or by `Match` statement.
    If `Hostname` is not defined in ssh config - is set to  `name`.
    """

    def __init__(
        self, name: str, hostname: str | None = None, user: str | None = None
    ) -> None:
        slots = {
            HOST_NAME_KEY: name,
            HOST_ADDRESS_KEY: hostname or name,
            HOST_SERVICE_NAME_KEY: "ssh",
        }
        if user:
            slots[HOST_SERVICE_USER_KEY] = user

        HostLeaf.__init__(self, slots, name)

    def get_description(self) -> str:
        host = str(self[HOST_ADDRESS_KEY])
        with suppress(KeyError):
            if user := self[HOST_SERVICE_USER_KEY]:
                host = f"{user}@{host}"

        return _("SSH host: %s") % host

    def get_gicon(self):
        return icons.ComposedIconSmall(self.get_icon_name(), "terminal")

    def get_text_representation(self) -> str:
        host = str(self[HOST_ADDRESS_KEY])
        with suppress(KeyError):
            if user := self[HOST_SERVICE_USER_KEY]:
                host = f"{user}@{host}"

        return f"ssh://{host}"

    def get_urilist_representation(self) -> list[str]:
        return [self.get_text_representation()]


class SSHConnect(Action):
    """Used to launch a terminal connecting to the specified SSH host.
    HOST_NAME_KEY is used for making connections.
    """

    def __init__(self):
        Action.__init__(self, name=_("Connect"))
        self.kupfer_add_alias("scp")

    def activate(self, leaf, iobj=None, ctx=None):
        host = leaf[HOST_NAME_KEY]
        with suppress(KeyError):
            if user := leaf[HOST_SERVICE_USER_KEY]:
                host = f"{user}@{host}"

        launch.spawn_in_terminal(["ssh", host])

    def get_description(self):
        return _("Connect to SSH host")

    def get_icon_name(self):
        return "network-server"

    def item_types(self):
        yield HostLeaf

    def valid_for_item(self, leaf):
        if leaf.check_key(HOST_SERVICE_NAME_KEY):
            return leaf[HOST_SERVICE_NAME_KEY] == "ssh"

        return False


class ScpFile(Action):
    """Send file or directory to remote host via scp."""

    def __init__(self):
        super().__init__(name=_("Send file to..."))

    def activate(self, leaf, iobj=None, ctx=None):
        assert leaf and isinstance(leaf, FileLeaf)
        assert iobj and isinstance(iobj, SSHLeaf)

        host = iobj[HOST_NAME_KEY]
        with suppress(KeyError):
            if user := iobj[HOST_SERVICE_USER_KEY]:
                host = f"{user}@{host}"

        cmd = ["scp"]
        if leaf.is_dir():
            cmd += ["-r"]

        cmd += [leaf.object, f"{host}:"]
        launch.spawn_in_terminal(cmd)

    def valid_for_item(self, leaf):
        return os.access(leaf.object, os.R_OK)

    def requires_object(self):
        return True

    def item_types(self):
        yield FileLeaf

    def object_types(self):
        yield SSHLeaf

    def get_description(self):
        return _("Copy file or directory to remote host using scp")

    def get_icon_name(self):
        return "go-next"


def _parse_host_stms(args: list[str]) -> ty.Iterator[tuple[str, str | None]]:
    """Load names from `Host` statement. Ignore names with `*` and exclusions."""
    for host in args:
        if "*" in host or host[0] == "!":
            continue

        user, _, host = host.partition("@")  # noqa:PLW2901
        if user and host:
            yield host, user
        else:
            yield user, None


def _parse_match_stmt(args: list[str]) -> tuple[str, str | None] | None:
    """Parse `Match` arguments and return tuple of (host, optional user)
    or None.

    Load only `User` and `Host` keys from this statement.
    """
    user = None
    host = None
    for key, val in itertools.pairwise(args):
        if "*" in val or val[0] == "!":
            continue

        key = key.lower()  # noqa:PLW2901
        if key == "host":
            host = val
        elif key == "user":
            user = val

    if not host:
        return None

    return host, user


class SSHSource(ToplevelGroupingSource, FilesystemWatchMixin):
    """Reads ~/.ssh/config and creates leaves for the hosts found."""

    source_scan_interval: int = 3600

    _ssh_home = os.path.expanduser("~/.ssh")
    _config_path = os.path.join(_ssh_home, "config")

    def __init__(self, name=_("SSH Hosts")):
        ToplevelGroupingSource.__init__(self, name, "hosts")
        self._version = 2
        self.monitor_token = None

    def initialize(self):
        ToplevelGroupingSource.initialize(self)
        self.monitor_token = self.monitor_files(self._config_path)

    def monitor_include_file(self, gfile):
        return gfile.get_path() in (self._config_path, self._ssh_home)

    def _get_items(self) -> ty.Iterable[SSHLeaf]:
        with open(self._config_path, encoding="UTF-8") as cfile:
            current_hosts: list[tuple[str, str | None]] = []
            current_hostname: str | None = None
            for line in cfile:
                line = line.strip()  # noqa:PLW2901
                if not line:
                    continue

                head, *args = line.split()
                head = head.lower()
                if head in ("host", "match"):
                    # new restriction, flush current data
                    for host, user in current_hosts:
                        yield SSHLeaf(host, current_hostname, user)

                    current_hosts.clear()
                    current_hostname = None

                    if head == "host":
                        # process only 'host' restriction; skip wildcard and
                        # negative entries
                        current_hosts = list(_parse_host_stms(args))
                    elif len(args) > 1 and (
                        hostuser := _parse_match_stmt(args)
                    ):
                        # process "key val" parameters of match
                        current_hosts = [hostuser]

                elif head == "hostname" and args:
                    # if found hostname use is as HOST_ADDRESS_KEY
                    current_hostname = args[0]

            for host, user in current_hosts:
                yield SSHLeaf(host, current_hostname, user)

    def get_items(self):
        try:
            return list(self._get_items())
        except OSError as exc:
            self.output_error(exc)
        except UnicodeError as exc:
            self.output_error(
                f"File {self._config_path} not in expected encoding (UTF-8)"
            )
            self.output_error(exc)

        return ()

    def get_description(self):
        return _("SSH hosts as specified in ~/.ssh/config")

    def get_icon_name(self):
        return "network-workgroup"

    def provides(self):
        yield SSHLeaf



"""
Changes:
    2012-10-17 Karol Będkowski:
        + control rhythmbox via dbus interface
        + load songs via dbus interface
    2023-02-19 KB:
        + catch errors when no mpris is available via dbus
        + simplify dbus string conversion

NOTE: this require Rhythmbox with mpris support (i.e. rhythmbox-plugins package
installed)
"""

from __future__ import annotations

__kupfer_name__ = _("Rhythmbox")
__kupfer_sources__ = ("RhythmboxSource",)
__description__ = _("Play and enqueue tracks and browse the music library")
__version__ = "2023.1"
__author__ = "US, Karol Będkowski"


import itertools
import operator
import os
import typing as ty
from collections import defaultdict
from hashlib import md5

import dbus
from gi.repository import Gio

from kupfer import config, icons, launch, plugin_support
from kupfer.obj import (
    Action,
    FileLeaf,
    Leaf,
    NotAvailableError,
    OperationError,
    RunnableLeaf,
    Source,
    SourceLeaf,
)
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.helplib import PicklingHelperMixin
from kupfer.plugin import rhythmbox_support
from kupfer.support import kupferstring, pretty, weaklib

if ty.TYPE_CHECKING:
    from gettext import gettext as _


plugin_support.check_dbus_connection()
plugin_support.check_command_available("rhythmbox-client")

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "toplevel_artists",
        "label": _("Include artists in top level"),
        "type": bool,
        "value": True,
    },
    {
        "key": "toplevel_albums",
        "label": _("Include albums in top level"),
        "type": bool,
        "value": False,
    },
    {
        "key": "toplevel_songs",
        "label": _("Include songs in top level"),
        "type": bool,
        "value": False,
    },
)

_BUS_NAME = "org.gnome.Rhythmbox3"
_OBJ_PATH_MPRIS = "/org/mpris/MediaPlayer2"
_OBJ_NAME_MPRIS_PLAYER = "org.mpris.MediaPlayer2.Player"
_OBJ_PATH_MEDIASERVC_ALL = "/org/gnome/UPnP/MediaServer2/Library/all"
_OBJ_NAME_MEDIA_CONT = "org.gnome.UPnP.MediaContainer2"


def _toutf8_lossy(ustr):
    return ustr.encode("UTF-8", "replace")


def _create_dbus_connection_mpris(obj_name, obj_path, activate=False):
    """Create dbus connection to Rhytmbox
    @activate: if True, start program if not running
    """
    interface = None
    sbus = dbus.SessionBus()
    try:
        proxy_obj = sbus.get_object(
            "org.freedesktop.DBus", "/org/freedesktop/DBus"
        )
        dbus_iface = dbus.Interface(proxy_obj, "org.freedesktop.DBus")
        if (activate or dbus_iface.NameHasOwner(_BUS_NAME)) and (
            obj := sbus.get_object(_BUS_NAME, obj_path)
        ):
            interface = dbus.Interface(obj, obj_name)

    except dbus.exceptions.DBusException as err:
        pretty.print_debug(err)

    return interface


def _tracknr(string: ty.Any) -> int | None:
    try:
        return int(string)
    except ValueError:
        return None


def _get_all_songs_via_dbus():
    try:
        iface = _create_dbus_connection_mpris(
            _OBJ_NAME_MEDIA_CONT, _OBJ_PATH_MEDIASERVC_ALL
        )
        if iface:
            for item in iface.ListItems(0, 9999, ["*"]):
                yield {
                    "album": str(item["Album"]),
                    "artist": str(item["Artist"]),
                    "title": str(item["DisplayName"]),
                    "track-number": _tracknr(item["TrackNumber"]),
                    "location": str(item["URLs"][0]),
                    "date": str(item["Date"]),
                }
    except Exception:
        pretty.print_exc(__name__, "_get_all_songs_via_dbus error")


def spawn_async(argv):
    try:
        launch.spawn_async_raise(argv)
    except launch.SpawnError as exc:
        raise OperationError(exc) from exc


def enqueue_songs(info, clear_queue=False, play_first=False):
    songs = list(info)
    if not songs:
        return

    qargv = ["rhythmbox-client"]
    if clear_queue:
        qargv.append("--clear-queue")

    if play_first and songs:
        song = songs[0]
        songs = songs[1:]
        uri = song["location"]
        qargv.extend(("--play-uri", uri))

    for song in songs:
        uri = song["location"]
        qargv.extend(("--enqueue", uri))

    spawn_async(qargv)


class ClearQueue(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Clear Queue"))

    def run(self, ctx=None):
        spawn_async(("rhythmbox-client", "--no-start", "--clear-queue"))

    def get_icon_name(self):
        return "edit-clear"


def _songs_from_leaf(leaf):
    "return a sequence of songs from @leaf"
    if isinstance(leaf, SongLeaf):
        return (leaf.object,)

    if isinstance(leaf, TrackCollection):
        return tuple(leaf.object)

    return ()


class PlayTracks(Action):
    action_accelerator = "o"

    rank_adjust = 5

    def __init__(self):
        Action.__init__(self, _("Play"))

    def activate(self, leaf, iobj=None, ctx=None):
        self.activate_multiple((leaf,))

    def activate_multiple(self, objects):
        # for multiple dispatch, play the first and enqueue the rest
        to_enqueue = []
        for leaf in objects:
            to_enqueue.extend(_songs_from_leaf(leaf))

        if to_enqueue:
            enqueue_songs(
                to_enqueue, clear_queue=len(to_enqueue) > 1, play_first=True
            )

    def get_description(self):
        return _("Play tracks in Rhythmbox")

    def get_icon_name(self):
        return "media-playback-start"


class Enqueue(Action):
    action_accelerator = "e"

    def __init__(self):
        Action.__init__(self, _("Enqueue"))

    def activate(self, leaf, iobj=None, ctx=None):
        self.activate_multiple((leaf,))

    def activate_multiple(self, objects):
        to_enqueue = []
        for leaf in objects:
            to_enqueue.extend(_songs_from_leaf(leaf))

        enqueue_songs(to_enqueue)

    def get_description(self):
        return _("Add tracks to the play queue")

    def get_gicon(self):
        return icons.ComposedIcon("gtk-execute", "media-playback-start")

    def get_icon_name(self):
        return "media-playback-start"


class SongLeaf(Leaf):
    serializable = 1

    def __init__(self, info, name=None):
        """Init with song info
        @info: Song information dictionary
        """
        Leaf.__init__(self, info, name or info["title"])

    def repr_key(self):
        """To distinguish songs by the same name"""
        return (
            self.object["title"],
            self.object["artist"],
            self.object["album"],
        )

    def get_actions(self):
        yield PlayTracks()
        yield Enqueue()
        yield GetFile()

    def get_description(self):
        # TRANS: Song description
        return _("by %(artist)s from %(album)s") % {
            "artist": self.object["artist"],
            "album": self.object["album"],
        }

    def get_icon_name(self):
        return "audio-x-generic"

    def get_text_representation(self):
        return self.name


class GetFile(Action):
    def __init__(self):
        super().__init__(_("Get File"))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        gfile = Gio.File.new_for_uri(leaf.object["location"])
        try:
            path = gfile.get_path()
        except Exception as exc:
            # On utf-8 decode error
            # FIXME: Unrepresentable path
            raise OperationError(exc) from exc

        if path:
            result = FileLeaf(path)
            if result.is_valid():
                return result

        raise NotAvailableError(str(leaf))


class CollectionSource(Source):
    def __init__(self, leaf):
        Source.__init__(self, str(leaf))
        self.leaf = leaf

    def get_items(self):
        for song in self.leaf.object:
            yield SongLeaf(song)

    def repr_key(self):
        return self.leaf.repr_key()

    def get_description(self):
        return self.leaf.get_description()

    def get_thumbnail(self, width, height):
        return self.leaf.get_thumbnail(width, height)

    def get_gicon(self):
        return self.leaf.get_gicon()

    def get_icon_name(self):
        return self.leaf.get_icon_name()


class TrackCollection(Leaf):
    """A generic track collection leaf, such as one for
    an Album or an Artist
    """

    def __init__(self, info, name):
        """Init with track collection
        @info: Should be a sequence of song information dictionaries
        """
        Leaf.__init__(self, info, name)

    def get_actions(self):
        yield PlayTracks()
        yield Enqueue()

    def has_content(self):
        return True

    def content_source(self, alternate=False):
        return CollectionSource(self)

    def get_icon_name(self):
        return "media-optical"


class AlbumLeaf(TrackCollection):
    def get_description(self):
        artist = None
        for song in self.object:
            if not artist:
                artist = song["artist"]
            elif artist != song["artist"]:
                # TRANS: Multiple artist description "Artist1 et. al. "
                artist = _("%s et. al.") % artist
                break
        # TRANS: Album description "by Artist"
        return _("by %s") % (artist,)

    def _get_thumb_local(self):
        # try local filesystem
        uri = self.object[0]["location"]
        artist = self.object[0]["artist"].lower()
        album = self.object[0]["album"].lower()
        gfile = Gio.File.new_for_uri(uri)
        cdir = gfile.resolve_relative_path("../").get_path()
        # We don't support unicode ATM
        bs_artist_album = " - ".join([artist, album])
        bs_artist_album2 = "-".join([artist, album])
        # " - ".join([us.encode("ascii", "ignore") for us in (artist, album)])
        cover_names = (
            "cover.jpg",
            "album.jpg",
            "albumart.jpg",
            "cover.gif",
            "album.png",
            ".folder.jpg",
            "folder.jpg",
            bs_artist_album + ".jpg",
            bs_artist_album2 + ".jpg",
        )
        try:
            for cover_name in os.listdir(cdir):
                if cover_name.lower() in cover_names:
                    cfile = gfile.resolve_relative_path("../" + cover_name)
                    return cfile.get_path()

        except OSError:
            pretty.print_exc(__name__)

        return None

    def _get_thumb_mediaart(self):
        """old thumb location"""
        ltitle = str(self).lower()
        # ignore the track artist -- use the space fallback
        # hash of ' ' as fallback
        hspace = "7215ee9c7d9dc229d2921a40e899ec5f"
        htitle = md5(_toutf8_lossy(ltitle)).hexdigest()
        hartist = hspace
        cache_name = f"album-{hartist}-{htitle}.jpeg"
        return config.get_cache_file(("media-art", cache_name))

    def get_thumbnail(self, width, height):
        if not hasattr(self, "cover_file"):
            # pylint: disable=attribute-defined-outside-init
            self.cover_file = (
                self._get_thumb_mediaart() or self._get_thumb_local()
            )

        return icons.get_pixbuf_from_file(self.cover_file, width, height)


class ArtistAlbumsSource(CollectionSource):
    def get_items(self):
        albums: dict[str, list[dict[str, ty.Any]]] = defaultdict(list)
        for song in self.leaf.object:
            albums[song["album"]].append(song)

        names = kupferstring.locale_sort(albums.keys())
        names.sort(key=lambda name: albums[name][0]["date"])
        for album in names:
            yield AlbumLeaf(albums[album], album)

    def should_sort_lexically(self):
        return False


class ArtistLeaf(TrackCollection):
    def get_description(self):
        # TRANS: Artist songs collection description
        return _("Tracks by %s") % (str(self),)

    def get_gicon(self):
        return icons.ComposedIcon("media-optical", "system-users")

    def content_source(self, alternate=False):
        if alternate:
            return CollectionSource(self)

        return ArtistAlbumsSource(self)


class RhythmboxAlbumsSource(Source):
    def __init__(self, library):
        Source.__init__(self, _("Albums"))
        self.library = library

    def get_items(self):
        for album in self.library:
            yield AlbumLeaf(self.library[album], album)

    def should_sort_lexically(self):
        return True

    def get_description(self):
        return _("Music albums in Rhythmbox Library")

    def get_gicon(self):
        return icons.ComposedIcon(
            "rhythmbox", "media-optical", emblem_is_fallback=True
        )

    def get_icon_name(self):
        return "rhythmbox"

    def provides(self):
        yield AlbumLeaf


class RhythmboxArtistsSource(Source):
    def __init__(self, library):
        Source.__init__(self, _("Artists"))
        self.library = library

    def get_items(self):
        for artist in self.library:
            yield ArtistLeaf(self.library[artist], artist)

    def should_sort_lexically(self):
        return True

    def get_description(self):
        return _("Music artists in Rhythmbox Library")

    def get_gicon(self):
        return icons.ComposedIcon(
            "rhythmbox", "system-users", emblem_is_fallback=True
        )

    def get_icon_name(self):
        return "rhythmbox"

    def provides(self):
        yield ArtistLeaf


def _locale_sort_artist_album_songs(artists):
    """Locale sort dictionary @artists by Artist, then Album;
    each artist in @artists should already contain songs
    grouped by album and sorted by track number.
    """
    for artist in kupferstring.locale_sort(artists):
        artist_songs = artists[artist]
        albums: dict[str, list[rhythmbox_support.Song]] = defaultdict(list)
        for album, songs in itertools.groupby(
            artist_songs, operator.itemgetter("album")
        ):
            albums[album].extend(songs)

        for album in kupferstring.locale_sort(albums):
            yield from albums[album]


class RhythmboxSongsSource(Source):
    """The whole song library in Leaf representation"""

    def __init__(self, library):
        Source.__init__(self, _("Songs"))
        self.library = library

    def get_items(self):
        for song in _locale_sort_artist_album_songs(self.library):
            yield SongLeaf(song)

    def get_actions(self):
        return ()

    def get_description(self):
        return _("Songs in Rhythmbox library")

    def get_gicon(self):
        return icons.ComposedIcon(
            "rhythmbox", "audio-x-generic", emblem_is_fallback=True
        )

    def provides(self):
        yield SongLeaf


class RhythmboxSource(AppLeafContentMixin, Source, PicklingHelperMixin):
    appleaf_content_id = ("rhythmbox", "org.gnome.Rhythmbox3")
    source_scan_interval: int = 3600

    def __init__(self):
        super().__init__(_("Rhythmbox"))
        self._version = 3
        self._songs = []

    def initialize(self):
        bus = dbus.SessionBus()
        weaklib.dbus_signal_connect_weakly(
            bus,
            "NameOwnerChanged",
            self._name_owner_changed,
            dbus_interface="org.freedesktop.DBus",
            arg0=_BUS_NAME,
        )

    def _name_owner_changed(self, name, old, new):
        if new:
            self.mark_for_update()

    def pickle_prepare(self):
        self.mark_for_update()

    def get_items(self):
        # first try to load songs via dbus
        songs = list(_get_all_songs_via_dbus())
        self._songs = songs = songs or self._songs
        albums = rhythmbox_support.parse_rhythmbox_albums(songs)
        artists = rhythmbox_support.parse_rhythmbox_artists(songs)
        yield ClearQueue()
        artist_source = RhythmboxArtistsSource(artists)
        album_source = RhythmboxAlbumsSource(albums)
        songs_source = RhythmboxSongsSource(artists)
        yield SourceLeaf(artist_source)
        yield SourceLeaf(album_source)
        yield SourceLeaf(songs_source)
        # we use get_leaves here to get sorting etc right
        if __kupfer_settings__["toplevel_artists"]:
            yield from artist_source.get_leaves()

        if __kupfer_settings__["toplevel_albums"]:
            yield from album_source.get_leaves()

        if __kupfer_settings__["toplevel_songs"]:
            yield from songs_source.get_leaves()

    def get_description(self):
        return _("Play and enqueue tracks and browse the music library")

    def get_icon_name(self):
        return "rhythmbox"

    def provides(self):
        yield RunnableLeaf
        yield SourceLeaf
        yield SongLeaf



__kupfer_name__ = _("Calculator")
__kupfer_actions__ = ("Calculate",)
__description__ = _("Calculate mathematical expressions")
__version__ = "2017.1"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import cmath
import inspect
import math
import textwrap
import typing as ty
from contextlib import suppress

from kupfer import launch, version
from kupfer.obj import Action, TextLeaf
from kupfer.support import pretty
from kupfer.ui import uiutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class IgnoreResultError(Exception):
    pass


class KupferSurprise(float):
    """kupfer

    cleverness to the inf**inf
    """

    def __call__(self, *args):
        launch.show_url(version.WEBSITE)
        raise IgnoreResultError


# pylint: disable=too-few-public-methods
class DummyResult:
    def __str__(self):
        return "<Result of last expression>"


class Help:
    """help()

    Show help about the calculator
    """

    def __call__(self):
        environment = make_environment(last_result=DummyResult())
        docstrings = []
        for attr in sorted(environment):
            if attr != "_" and attr.startswith("_"):
                continue

            val = environment[attr]
            if not callable(val):
                docstrings.append(f"{attr} = {val}")
                continue

            with suppress(AttributeError):
                try:
                    # use .replace() to remove unimportant '/' marker in signature
                    sig = str(inspect.signature(val)).replace(", /)", ")")
                    doc = f"{attr}{sig}\n{val.__doc__}"
                except ValueError:
                    doc = val.__doc__

                docstrings.append(doc)

        formatted = []
        maxlen = 72
        left_margin = 4
        for docstr in docstrings:
            # Wrap the description and align continued lines
            docsplit = docstr.split("\n", 1)
            if len(docsplit) <= 1:
                formatted.append(docstr)
                continue

            wrapped_lines = textwrap.wrap(
                docsplit[1].strip(), maxlen - left_margin
            )
            wrapped = ("\n" + " " * left_margin).join(wrapped_lines)
            formatted.append(f"{docsplit[0]}\n    {wrapped}")

        uiutils.show_text_result("\n\n".join(formatted), _("Calculator"))
        raise IgnoreResultError

    def __complex__(self):
        return self()


def make_environment(last_result=None):
    "Return a namespace for the calculator's expressions to be executed in."
    environment = dict(vars(math))
    environment.update(vars(cmath))
    # define some constants missing
    if last_result is not None:
        environment["_"] = last_result

    environment["help"] = Help()
    environment["kupfer"] = KupferSurprise("inf")
    # make the builtins inaccessible
    environment["__builtins__"] = {}
    return environment


def format_result(res):
    cres = complex(res)
    parts = []
    if cres.real:  # pylint: disable=using-constant-test
        parts.append(str(cres.real))

    if cres.imag:  # pylint: disable=using-constant-test
        parts.append(str(complex(0, cres.imag)))

    return "+".join(parts) or str(res)


class Calculate(Action):
    # since it applies only to special queries, we can up the rank
    rank_adjust = 10
    # global last_result
    last_result: ty.ClassVar[dict[str, ty.Any]] = {"last": None}

    def __init__(self):
        Action.__init__(self, _("Calculate"))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        expr = leaf.object.lstrip("= ")

        # try to add missing parentheses
        brackets_missing = expr.count("(") - expr.count(")")
        if brackets_missing > 0:
            expr += ")" * brackets_missing

        # hack: change all decimal points (according to current locale) to '.'
        # expr = expr.replace(locale.localeconv()['decimal_point'], '.')
        environment = make_environment(self.last_result["last"])
        pretty.print_debug(__name__, "Evaluating", expr)
        try:
            result = eval(expr, environment)  # pylint: disable=eval-used
            resultstr = format_result(result)
            self.last_result["last"] = result
        except IgnoreResultError:
            return None
        except Exception as exc:
            pretty.print_error(__name__, type(exc).__name__, exc)
            resultstr = str(exc)

        return TextLeaf(resultstr)

    def item_types(self):
        yield TextLeaf

    def valid_for_item(self, leaf):
        text = leaf.object
        return bool(text) and text.startswith("=")

    def get_description(self):
        return None



"""
This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import functools
import inspect
import typing as ty
from collections import OrderedDict

__all__ = ("LruCache", "evaluate_once", "simple_cache")

K = ty.TypeVar("K")
V = ty.TypeVar("V")
VD = ty.TypeVar("VD")


def _get_point_of_create(offset: int = 2) -> str:
    """Find information about function caller with `offset` in tracestack."""
    for frame in inspect.stack()[offset:]:
        if "kupfer" in frame.filename:
            return f"{frame.filename}:{frame.lineno}"

    return "?"


_sentinel = object()


class LruCache(ty.Generic[K, V]):
    """Least-recently-used cache mapping of size *maxsize*.

    *name* is optional cache name for debug purpose. If not given, is place
    when cache is created (filename:lineno).
    """

    def __init__(self, maxsize: int, name: str | None = None) -> None:
        super().__init__()
        self._data: OrderedDict[K, V] = OrderedDict()
        self._maxsize = maxsize
        self._name = name or _get_point_of_create()
        self._hit = 0
        self._miss = 0
        self._inserts = 0

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> ty.Iterator[K]:
        return self._data.__iter__()

    def __setitem__(self, key: K, value: V) -> None:
        self._inserts += 1
        # set add item on the end of dict
        data = self._data
        if key in self._data:
            # check is item already in dict by trying to move it to the end
            data.move_to_end(key, last=True)
        else:
            # item not found in dict so add it
            data[key] = value

            if len(data) > self._maxsize:
                # remove the first item (was inserted longest time ago)
                data.popitem(last=False)

    def __getitem__(self, key: K) -> V:
        # try to get item from dict, if not found KeyError is raised
        try:
            value = self._data[key]
        except KeyError:
            self._miss += 1
            raise

        self._hit += 1
        # found, so move it to the end
        self._data.move_to_end(key, last=True)
        return value

    def get(self, key: K, default: V | VD | None = None) -> V | VD | None:
        try:
            return self[key]
        except KeyError:
            return default

    def keys(self) -> ty.KeysView[K]:
        return self._data.keys()

    def values(self) -> ty.ValuesView[V]:
        return self._data.values()

    def items(self) -> ty.ItemsView[K, V]:
        return self._data.items()

    def clear(self) -> None:
        self._data.clear()

    def __str__(self) -> str:
        return (
            f"<LruCache '{self._name}': maxsize={self._maxsize}, "
            f"items={len(self)}, hit={self._hit}, miss={self._miss}, "
            f"inserts={self._inserts}>"
        )

    def get_or_insert(self, key: K, creator: ty.Callable[[], V]) -> V:
        """Get value from cache. If not exists - create with with `creator`
        function and insert into cache."""
        try:
            val = self._data[key]
        except KeyError:
            self._miss += 1
            self._inserts += 1
            val = self._data[key] = creator()
        else:
            self._hit += 1

        return val


RT = ty.TypeVar("RT")  # return type


class simple_cache(ty.Generic[RT]):  # pylint: disable=invalid-name # noqa:N801
    """Function wrapper that remember (cache) one result and return it
    if no arguments changed.

    Wrapper properties:
        *cache_current_args*   remembered (last) arguments
        *cache_current_value*  remembered (last) function result
        *cache_hit*            number of result returned from cache
        *cache_miss*           number of wrapped function calls

    Usage::

        @SimpleCache
        def function(args):
            ...
    """

    cache_current_value: RT
    cache_current_args: ty.Any
    cache_hit: int
    cache_miss: int

    def __init__(self, func: ty.Callable[..., RT]):
        """Create SimpleCache for *func* function."""
        functools.update_wrapper(self, func)
        self.func = func
        self.cache_clear()
        self.name = _get_point_of_create()

    def __call__(self, *args: ty.Any, **kwargs: ty.Any) -> RT:
        if self.cache_current_args == (args, kwargs):
            self.cache_hit += 1
            return self.cache_current_value

        result = self.func(*args, **kwargs)

        self.cache_current_args = (args, kwargs)
        self.cache_current_value = result
        self.cache_miss += 1

        return result

    def cache_clear(self) -> None:
        """Clear cache and stats."""
        self.cache_current_value = None  # type: ignore
        self.cache_current_args = None
        self.cache_hit = 0
        self.cache_miss = 0

    def __str__(self) -> str:
        return (
            f"<simple_cache '{self.name}': hit={self.cache_hit},"
            f" miss={self.cache_miss}>"
        )


def evaluate_once(func: ty.Callable[..., RT]) -> ty.Callable[..., RT]:
    """Decorator that run wrapped function once and always return computed
    value. No thread safe."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = getattr(wrapper, "cached_value", _sentinel)
        if result is not _sentinel:
            return result

        result = func(*args, **kwargs)
        wrapper.cached_value = result  # type: ignore
        return result

    return wrapper



# Copyright © 2022 Karol Będkowski <Karol Będkowski@kkomp>
#
# Distributed under terms of the GPLv3 license.

"""
Plugin allow control domains (virtual machines) managed by libvirt.
"""

from __future__ import annotations

__kupfer_name__ = _("Libvirt")
__kupfer_sources__ = ("LibvirtDomainsSource",)
__description__ = _("Control libvirt guest domains.")
__version__ = "0.1"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import typing as ty

import libvirt

from kupfer import icons, launch, plugin_support
from kupfer.obj import Action, Leaf, Source
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "connection",
        "label": _("Connection"),
        "type": str,
        "value": "qemu:///system",
    },
)


class Domain(Leaf):
    def __init__(self, obj, name, description):
        Leaf.__init__(self, obj, name)
        self.description = description

    def get_description(self):
        return self.description

    def get_actions(self):
        conn = _ConnManager.instance().get_conn()
        if not conn:
            return

        domain = conn.lookupByName(self.object)
        state, _reason = domain.state()

        # https://libvirt.org/html/libvirt-libvirt-domain.html#virDomainState
        if state == libvirt.VIR_DOMAIN_RUNNING:
            yield DomAction(_("Pause"), "pause", "pause")
            yield DomAction(_("Reboot"), "system-reboot", "reboot", -5)
            yield DomAction(
                _("Send Power Off Signal"),
                "system-shutdown",
                "poweroff-signal",
                -10,
            )
            yield DomAction(_("Power Off"), "system-shutdown", "poweroff", -15)
        elif state in (
            libvirt.VIR_DOMAIN_PAUSED,
            libvirt.VIR_DOMAIN_PMSUSPENDED,
        ):
            yield DomAction(_("Resume"), "system-run", "resume")
        elif state in (
            libvirt.VIR_DOMAIN_BLOCKED,
            libvirt.VIR_DOMAIN_SHUTDOWN,
        ):
            pass
        else:  # libvirt.VIR_DOMAIN_SHUTOFF:
            yield DomAction(_("Power On"), "system-run", "start")

        yield OpenConsoleAction()

    def get_gicon(self):
        return icons.ComposedIconSmall("computer", "virt-manager")


class DomAction(Action):
    def __init__(self, name, icon, command, rank_adjust=0):
        Action.__init__(self, name)
        self._icon = icon
        self.rank_adjust = rank_adjust
        self.command = command

    def get_icon_name(self):
        return self._icon

    def item_types(self):
        yield Domain

    def activate(self, leaf, iobj=None, ctx=None):
        with libvirt.open(__kupfer_settings__["connection"] or None) as conn:
            domain = conn.lookupByName(leaf.object)

            if self.command == "poweroff":
                domain.shutdown()
            elif self.command == "pause":
                domain.suspend()
            elif self.command == "reboot":
                domain.reboot(libvirt.VIR_DOMAIN_REBOOT_DEFAULT)
            elif self.command == "resume":
                domain.resume()
            elif self.command == "start":
                domain.create()
            elif self.command == "poweroff-signal":
                domain.shutdownFlags(libvirt.VIR_DOMAIN_SHUTDOWN_DEFAULT)


def _get_domain_metadata(domain, key):
    try:
        return domain.metadata(key, None)
    except libvirt.libvirtError:
        pass
    return None


class _ConnManager(pretty.OutputMixin):
    """Libvirt connection manager. Only for read-only connections."""

    _instance: _ConnManager | None = None

    @classmethod
    def instance(cls) -> _ConnManager:
        if not cls._instance:
            cls._instance = _ConnManager()

        return cls._instance

    def __init__(self):
        self.conn: libvirt.virConnect | None = None
        libvirt.virEventRegisterDefaultImpl()

    def close(self):
        if self.conn:
            self.conn.close()
            self.conn = None

    def get_conn(self) -> libvirt.virConnect | None:
        if self.conn is None:
            self._open()

        return self.conn

    def _open(self) -> None:
        connstr = __kupfer_settings__["connection"] or None
        self.output_debug("LibVirt Connection open", connstr)
        self.conn = libvirt.openReadOnly(connstr)
        if self.conn:
            self.conn.setKeepAlive(5, 3)



class LibvirtDomainsSource(AppLeafContentMixin, Source):
    source_use_cache = False
    appleaf_content_id = "virt-manager"

    def __init__(self, name=_("Libvirt domains")):
        Source.__init__(self, name)
        self.cmgr: _ConnManager | None = None

    def initialize(self):
        # prevent logging errors from libvirt
        libvirt.registerErrorHandler(lambda _userdata, _err: None, ctx=None)
        __kupfer_settings__.connect(
            "plugin-setting-changed", self._on_setting_changed
        )
        self.cmgr = _ConnManager.instance()
        self._on_setting_changed("", "", "")


    def finalize(self):
        if self.cmgr:
            self.cmgr.close()
            self.cmgr = None

    def _callback(self, _conn, dom, event, detail, _opaque):
        pretty.print_debug(
            "LibvirtDomainsSource event", dom.name(), dom.ID(), event, detail
        )
        self.mark_for_update()

    def get_items(self):
        assert self.cmgr
        conn = self.cmgr.get_conn()
        if not conn:
            return

        for dom in conn.listAllDomains():
            name = dom.name()
            title = _get_domain_metadata(
                dom, libvirt.VIR_DOMAIN_METADATA_TITLE
            )
            descr = _get_domain_metadata(
                dom, libvirt.VIR_DOMAIN_METADATA_DESCRIPTION
            )
            yield Domain(name, title or name, descr)

    def get_description(self):
        return None

    def provides(self):
        yield Domain

    def should_sort_lexically(self):
        return True

    def _on_setting_changed(self, settings, key, value):
        assert self.cmgr

        self.cmgr.close()
        if conn := self.cmgr.get_conn():
            conn.domainEventRegister(self._callback, None)

        self.mark_for_update()


class OpenConsoleAction(Action):
    def __init__(self):
        Action.__init__(self, name=_("Open console"))

    def get_icon_name(self):
        return "virt-manager"

    def item_types(self):
        yield Domain

    def activate(self, leaf, iobj=None, ctx=None):
        conn = __kupfer_settings__["connection"] or "qemu:///system"
        domain = leaf.object
        launch.spawn_async(
            [
                "virt-manager",
                "--connect",
                conn,
                "--show-domain-console",
                domain,
            ]
        )

    def get_description(self):
        return _("Open Virtual Machine Manager console for domain")



__kupfer_name__ = _("File Actions")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_actions__ = ("MoveTo", "Rename", "CopyTo", "Edit")
__description__ = _("More file actions")
__version__ = "2023-02-25"
__author__ = "Ulrik, KB"

import os
import typing as ty

# since "path" is a very generic name, you often forget..
from os import path as os_path
from pathlib import Path

from gi.repository import Gio, GLib

from kupfer import launch
from kupfer.core import settings
from kupfer.obj import Action, FileLeaf, OperationError, TextLeaf, TextSource
from kupfer.support import pretty, task

if ty.TYPE_CHECKING:
    from gettext import gettext as _


def _good_destination(dpath, spath):
    """If directory path @dpath is a valid destination for file @spath
    to be copied or moved to.
    """
    if not os_path.isdir(dpath):
        return False

    spath = os_path.normpath(spath)
    dpath = os_path.normpath(dpath)
    if not os.access(dpath, os.R_OK | os.W_OK | os.X_OK):
        return False

    cpfx = os_path.commonprefix((spath, dpath))
    return not (os_path.samefile(dpath, spath) or cpfx == spath)


class MoveTo(Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Move To..."))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj

        sfile = leaf.get_gfile()
        bname = sfile.get_basename()
        dfile = iobj.get_gfile().get_child(bname)
        try:
            ret = sfile.move(
                dfile, Gio.FileCopyFlags.ALL_METADATA, None, None, None
            )
            self.output_debug(f"Move {sfile} to {dfile} (ret: {ret})")
        except GLib.Error as exc:
            raise OperationError(str(exc)) from exc

        return FileLeaf(dfile.get_path())

    def valid_for_item(self, leaf):
        return os.access(leaf.object, os.R_OK | os.W_OK)

    def requires_object(self):
        return True

    def item_types(self):
        yield FileLeaf

    def object_types(self):
        yield FileLeaf

    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)

    def get_description(self):
        return _("Move file to new location")

    def get_icon_name(self):
        return "go-next"


class RenameSource(TextSource):
    """A source for new names for a file;
    here we "autopropose" the source file's extension,
    but allow overriding it as well as renaming to without
    extension (either using a terminating space, or selecting the
    normal TextSource-returned string).
    """

    def __init__(self, sourcefile):
        self.sourcefile = sourcefile
        name = _("Rename To...").rstrip(".")
        TextSource.__init__(self, name)

    def get_rank(self):
        # this should rank high
        return 100

    def get_items(self, text):
        if not text:
            return
        basename = os_path.basename(self.sourcefile.object)
        _root, ext = os_path.splitext(basename)
        t_root, t_ext = os_path.splitext(text)
        if text.endswith(" "):
            yield TextLeaf(text.rstrip())
        else:
            yield TextLeaf(text) if t_ext else TextLeaf(t_root + ext)

    def get_gicon(self):
        return self.sourcefile.get_gicon()


class Rename(Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Rename To..."))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj

        sfile = leaf.get_gfile()
        dest = os_path.join(os_path.dirname(leaf.object), iobj.object)
        dfile = Gio.File.new_for_path(dest)
        try:
            ret = sfile.move(
                dfile, Gio.FileCopyFlags.ALL_METADATA, None, None, None
            )
            self.output_debug(f"Move {sfile} to {dfile} (ret: {ret})")
        except GLib.Error as exc:
            raise OperationError(str(exc)) from exc

        return FileLeaf(dfile.get_path())

    def activate_multiple(self, objs, iobjs):
        raise NotImplementedError

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return os.access(leaf.object, os.R_OK | os.W_OK)

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    def valid_object(self, obj, for_item):
        dest_dir = Path(for_item.object).parent
        return dest_dir.is_dir() and not dest_dir.joinpath(obj.object).exists()

    def object_source(self, for_item=None):
        assert for_item
        return RenameSource(for_item)

    def get_description(self):
        return None


class CopyTo(Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Copy To..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        assert ctx
        sfile = leaf.get_gfile()
        dfile = iobj.get_gfile().get_child(os_path.basename(leaf.object))
        return CopyTask(str(leaf), sfile, dfile, ctx)

    def is_async(self):
        return True

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return (not leaf.is_dir()) and os.access(leaf.object, os.R_OK)

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf

    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)

    def get_description(self):
        return _("Copy file to a chosen location")


class CopyTask(task.ThreadTask, pretty.OutputMixin):
    def __init__(self, name, gsource, gdest, ctx):
        super().__init__(name)
        self.gsource = gsource
        self.gdest = gdest
        self.ctx = ctx

    def thread_do(self):
        try:
            # FIXME: This should be async
            self.output_debug(f"Copy {self.gsource} to {self.gdest}")
            ret = self.gsource.copy(
                self.gdest, Gio.FileCopyFlags.ALL_METADATA, None, None, None
            )
            self.output_debug(f"Copy ret {ret!r}")
        except GLib.Error as exc:
            # pylint: disable=no-member
            raise OperationError(exc.message) from exc

    def thread_finish(self):
        self.ctx.register_late_result(FileLeaf(self.gdest.get_path()))

    def thread_finally(self, exc_info):
        if exc_info is not None:
            self.ctx.register_late_error(exc_info)


class Edit(Action):
    action_accelerator = "e"

    def __init__(self):
        super().__init__(_("Edit file content"))

    def activate(self, leaf, iobj=None, ctx=None):
        sett = settings.get_settings_controller()
        editor = sett.get_preferred_alternative("editor")
        if not editor:
            return

        argv: list[str] = editor["argv"]
        new_argv = list(_replace_argv_filename(argv, (leaf,)))
        if new_argv == argv:
            new_argv.append(leaf.object)

        if editor["terminal"]:
            launch.spawn_in_terminal(new_argv)
        else:
            launch.spawn_async(new_argv)

    def activate_multiple(self, objects):
        sett = settings.get_settings_controller()
        editor = sett.get_preferred_alternative("editor")
        if not editor:
            return

        argv: list[str] = editor["argv"]
        if "%f" in argv or len(objects) == 1:
            # editor not support probably multiple files at once
            for obj in objects:
                self.activate(obj)

            return

        new_argv = list(_replace_argv_filename(argv, objects))
        if new_argv == argv:
            new_argv.extend(obj.object for obj in objects)

        if editor["terminal"]:
            launch.spawn_in_terminal(new_argv)
        else:
            launch.spawn_async(new_argv)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return not leaf.is_dir()

    def get_icon_name(self):
        return "edit"

    def get_description(self):
        return _("Open file in configured editor")


def _replace_argv_filename(
    argv: list[str], leaves: ty.Iterable[FileLeaf]
) -> ty.Iterable[str]:
    for part in argv:
        if part in ("%F", "%f"):
            yield from (leaf.object for leaf in leaves)
        else:
            yield part



from __future__ import annotations

__kupfer_name__ = _("Triggers")
__kupfer_sources__ = ("Triggers",)
__kupfer_actions__ = ("AddTrigger",)
__description__ = _(
    "Assign global keybindings (triggers) to objects created "
    "with 'Compose Command'."
)
__version__ = "2009-12-30"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import typing as ty

from gi.repository import GLib, Gtk

from kupfer import plugin_support, puid
from kupfer.core import commandexec
from kupfer.obj import Action, OperationError, RunnableLeaf, Source
from kupfer.obj.compose import ComposedLeaf
from kupfer.support import task
from kupfer.ui import getkey_dialog, keybindings, uievents

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_keybinding_support()


class Trigger(RunnableLeaf):
    def get_actions(self):
        yield from RunnableLeaf.get_actions(self)
        yield RemoveTrigger()

    def wants_context(self):
        return True

    def is_valid(self):
        return Triggers.has_trigger(self.object)

    def run(self, ctx=None):
        assert ctx
        return Triggers.perform_trigger(ctx, self.object)

    def repr_key(self):
        return self.object


class Triggers(Source):
    instance: Triggers | None = None

    def __init__(self):
        Source.__init__(self, _("Triggers"))
        self.trigger_table = {}

    def config_save(self):
        return {"triggers": self.trigger_table, "version": self.version}

    def config_save_name(self):
        return __name__

    def config_restore(self, state):
        self.trigger_table = state["triggers"]
        return True

    def initialize(self):
        Triggers.instance = self
        keybindings.get_keybound_object().connect(
            "keybinding", self._on_keybinding_callback
        )
        for target, (keystr, _name, _id) in self.trigger_table.items():
            keybindings.bind_key(keystr, target)

        self.output_debug("Loaded triggers, count:", len(self.trigger_table))

    def finalize(self):
        for target in self.trigger_table:
            keybindings.bind_key(None, target)

    def _on_keybinding_callback(self, keyobj, target, display, event_time):
        if not self.has_trigger(target):
            return

        ui_ctx = uievents.gui_context_from_keyevent(event_time, display)
        ctx = commandexec.default_action_execution_context()
        exec_token = ctx.make_execution_token(ui_ctx)
        self.perform_trigger(exec_token, target)

    def get_items(self):
        for target, (keystr, name, _id) in self.trigger_table.items():
            label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(keystr))
            yield Trigger(target, f"{label or keystr} ({name})")

    def should_sort_lexically(self):
        return True

    def provides(self):
        yield Trigger

    @classmethod
    def has_trigger(cls, target):
        assert cls.instance
        return target in cls.instance.trigger_table

    @classmethod
    def perform_trigger(cls, ctx, target):
        assert cls.instance
        try:
            _, _, id_ = cls.instance.trigger_table[target]
        except KeyError as exc:
            raise OperationError(f"Trigger '{target}' does not exist") from exc

        if (obj := puid.resolve_unique_id(id_)) is not None:
            return obj.run(ctx)  # type:ignore

        return None

    @classmethod
    def add_trigger(cls, leaf, keystr):
        assert cls.instance
        # pylint: disable=protected-access
        cls.instance._add_trigger(leaf, keystr)  # noqa:SLF001

    @classmethod
    def remove_trigger(cls, target):
        assert cls.instance
        # pylint: disable=protected-access
        cls.instance._remove_trigger(target)  # noqa:SLF001

    def _add_trigger(self, leaf, keystr):
        for target in range(*keybindings.KEYRANGE_TRIGGERS):
            if target not in self.trigger_table:
                break

        keybindings.bind_key(keystr, target)
        name = str(leaf)
        self.trigger_table[target] = (keystr, name, puid.get_unique_id(leaf))
        self.mark_for_update()

    def _remove_trigger(self, target):
        self.trigger_table.pop(target, None)
        keybindings.bind_key(None, target)
        self.mark_for_update()

    def get_icon_name(self):
        return "key_bindings"


def _try_bind_key(keystr):
    label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(keystr))
    if label is None:
        return False

    if len(label) == 1 and label.isalnum():
        return False

    target = keybindings.KEYRANGE_TRIGGERS[-1] - 1
    if succ := keybindings.bind_key(keystr, target):
        keybindings.bind_key(None, target)

    return succ


class BindTask(task.Task):
    def __init__(self, leaf, screen):
        super().__init__()
        self.leaf = leaf
        self.screen = screen

    def start(self, finish_callback):
        GLib.idle_add(self.ask_key, finish_callback)

    def ask_key(self, finish_callback):
        keystr = getkey_dialog.ask_for_key(
            _try_bind_key, screen=self.screen, show_clear=False
        )
        if keystr:
            Triggers.add_trigger(self.leaf, keystr)

        finish_callback(self)


class AddTrigger(Action):
    def __init__(self):
        Action.__init__(self, _("Add Trigger..."))

    def is_async(self):
        return True

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        return BindTask(leaf, ctx.environment.get_screen())

    def item_types(self):
        yield ComposedLeaf

    def get_icon_name(self):
        return "list-add"


class RemoveTrigger(Action):
    def __init__(self):
        Action.__init__(self, _("Remove Trigger"))

    def activate(self, leaf, iobj=None, ctx=None):
        Triggers.remove_trigger(leaf.object)

    def get_icon_name(self):
        return "list-remove"
